# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE.txt in the project root for
# license information.
# -------------------------------------------------------------------------
from azure.core.pipeline.transport import (
    AsyncHttpResponse as PipelineTransportAsyncHttpResponse,
    AsyncHttpTransport,
    AioHttpTransport,
    HttpRequest,
    AioHttpTransportResponse,
)
from azure.core.pipeline.transport._aiohttp import AioHttpStreamDownloadGenerator
from azure.core.rest._http_response_impl_async import AsyncHttpResponseImpl as RestAsyncHttpResponse
from azure.core.pipeline.policies import HeadersPolicy
from azure.core.pipeline import AsyncPipeline
from azure.core.exceptions import HttpResponseError, ServiceResponseError
from utils import HTTP_REQUESTS, request_and_responses_product
import pytest
import sys
import aiohttp


# transport = mock.MagicMock(spec=AsyncHttpTransport)
# MagicMock support async cxt manager only after 3.8
# https://github.com/python/cpython/pull/9296


class MockAsyncHttpTransport(AsyncHttpTransport):
    async def __aenter__(self):
        return self

    async def __aexit__(self, *args):
        pass

    async def open(self):
        pass

    async def close(self):
        pass

    async def send(self, request, **kwargs):
        pass


class PipelineTransportMockResponse(PipelineTransportAsyncHttpResponse):
    def __init__(self, request, body, content_type):
        super().__init__(request, None)
        self._body = body
        self.content_type = content_type

    def body(self):
        return self._body


class RestMockResponse(RestAsyncHttpResponse):
    def __init__(self, request, body, content_type):
        super(RestMockResponse, self).__init__(
            request=request,
            internal_response=None,
            content_type=content_type,
            block_size=None,
            status_code=200,
            reason="OK",
            headers={},
            stream_download_generator=None,
        )
        # the impl takes in a lot more kwargs. It's not public and is a
        # helper implementation shared across our azure core transport responses
        self._content = body
        self._is_closed = True
        self._is_stream_consumed = True


MOCK_RESPONSES = [PipelineTransportMockResponse, RestMockResponse]


@pytest.mark.asyncio
@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
async def test_basic_options_aiohttp(port, http_request):
    request = http_request("OPTIONS", "http://localhost:{}/basic/string".format(port))
    async with AsyncPipeline(AioHttpTransport(), policies=[]) as pipeline:
        response = await pipeline.run(request)

    assert pipeline._transport.session is None
    assert isinstance(response.http_response.status_code, int)


@pytest.mark.asyncio
@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
async def test_multipart_send(http_request):
    transport = MockAsyncHttpTransport()

    class RequestPolicy(object):
        async def on_request(self, request):
            # type: (PipelineRequest) -> None
            request.http_request.headers["x-ms-date"] = "Thu, 14 Jun 2018 16:46:54 GMT"

    req0 = http_request("DELETE", "/container0/blob0")
    req1 = http_request("DELETE", "/container1/blob1")

    request = http_request("POST", "http://account.blob.core.windows.net/?comp=batch")
    request.set_multipart_mixed(
        req0,
        req1,
        policies=[RequestPolicy()],
        boundary="batch_357de4f7-6d0b-4e02-8cd2-6361411a9525",  # Fix it so test are deterministic
    )

    async with AsyncPipeline(transport) as pipeline:
        await pipeline.run(request)

    assert request.body == (
        b"--batch_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 0\r\n"
        b"\r\n"
        b"DELETE /container0/blob0 HTTP/1.1\r\n"
        b"x-ms-date: Thu, 14 Jun 2018 16:46:54 GMT\r\n"
        b"\r\n"
        b"\r\n"
        b"--batch_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 1\r\n"
        b"\r\n"
        b"DELETE /container1/blob1 HTTP/1.1\r\n"
        b"x-ms-date: Thu, 14 Jun 2018 16:46:54 GMT\r\n"
        b"\r\n"
        b"\r\n"
        b"--batch_357de4f7-6d0b-4e02-8cd2-6361411a9525--\r\n"
    )


@pytest.mark.asyncio
@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
async def test_multipart_send_with_context(http_request):
    transport = MockAsyncHttpTransport()
    header_policy = HeadersPolicy()

    class RequestPolicy(object):
        async def on_request(self, request):
            # type: (PipelineRequest) -> None
            request.http_request.headers["x-ms-date"] = "Thu, 14 Jun 2018 16:46:54 GMT"

    req0 = http_request("DELETE", "/container0/blob0")
    req1 = http_request("DELETE", "/container1/blob1")

    request = http_request("POST", "http://account.blob.core.windows.net/?comp=batch")
    request.set_multipart_mixed(
        req0,
        req1,
        policies=[header_policy, RequestPolicy()],
        boundary="batch_357de4f7-6d0b-4e02-8cd2-6361411a9525",  # Fix it so test are deterministic
        headers={"Accept": "application/json"},
    )

    async with AsyncPipeline(transport) as pipeline:
        await pipeline.run(request)

    assert request.body == (
        b"--batch_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 0\r\n"
        b"\r\n"
        b"DELETE /container0/blob0 HTTP/1.1\r\n"
        b"Accept: application/json\r\n"
        b"x-ms-date: Thu, 14 Jun 2018 16:46:54 GMT\r\n"
        b"\r\n"
        b"\r\n"
        b"--batch_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 1\r\n"
        b"\r\n"
        b"DELETE /container1/blob1 HTTP/1.1\r\n"
        b"Accept: application/json\r\n"
        b"x-ms-date: Thu, 14 Jun 2018 16:46:54 GMT\r\n"
        b"\r\n"
        b"\r\n"
        b"--batch_357de4f7-6d0b-4e02-8cd2-6361411a9525--\r\n"
    )


@pytest.mark.asyncio
@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
async def test_multipart_send_with_one_changeset(http_request):
    transport = MockAsyncHttpTransport()
    requests = [http_request("DELETE", "/container0/blob0"), http_request("DELETE", "/container1/blob1")]
    changeset = http_request("", "")
    changeset.set_multipart_mixed(*requests, boundary="changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525")

    request = http_request("POST", "http://account.blob.core.windows.net/?comp=batch")
    request.set_multipart_mixed(changeset, boundary="batch_357de4f7-6d0b-4e02-8cd2-6361411a9525")

    async with AsyncPipeline(transport) as pipeline:
        await pipeline.run(request)

    assert request.body == (
        b"--batch_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: multipart/mixed; boundary=changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"\r\n"
        b"--changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 0\r\n"
        b"\r\n"
        b"DELETE /container0/blob0 HTTP/1.1\r\n"
        b"\r\n"
        b"\r\n"
        b"--changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 1\r\n"
        b"\r\n"
        b"DELETE /container1/blob1 HTTP/1.1\r\n"
        b"\r\n"
        b"\r\n"
        b"--changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525--\r\n"
        b"\r\n"
        b"--batch_357de4f7-6d0b-4e02-8cd2-6361411a9525--\r\n"
    )


@pytest.mark.asyncio
@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
async def test_multipart_send_with_multiple_changesets(http_request):
    transport = MockAsyncHttpTransport()
    changeset1 = http_request("", "")
    changeset1.set_multipart_mixed(
        http_request("DELETE", "/container0/blob0"),
        http_request("DELETE", "/container1/blob1"),
        boundary="changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525",
    )
    changeset2 = http_request("", "")
    changeset2.set_multipart_mixed(
        http_request("DELETE", "/container2/blob2"),
        http_request("DELETE", "/container3/blob3"),
        boundary="changeset_8b9e487e-a353-4dcb-a6f4-0688191e0314",
    )

    request = http_request("POST", "http://account.blob.core.windows.net/?comp=batch")
    request.set_multipart_mixed(
        changeset1,
        changeset2,
        boundary="batch_357de4f7-6d0b-4e02-8cd2-6361411a9525",
    )

    async with AsyncPipeline(transport) as pipeline:
        await pipeline.run(request)

    assert request.body == (
        b"--batch_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: multipart/mixed; boundary=changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"\r\n"
        b"--changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 0\r\n"
        b"\r\n"
        b"DELETE /container0/blob0 HTTP/1.1\r\n"
        b"\r\n"
        b"\r\n"
        b"--changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 1\r\n"
        b"\r\n"
        b"DELETE /container1/blob1 HTTP/1.1\r\n"
        b"\r\n"
        b"\r\n"
        b"--changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525--\r\n"
        b"\r\n"
        b"--batch_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: multipart/mixed; boundary=changeset_8b9e487e-a353-4dcb-a6f4-0688191e0314\r\n"
        b"\r\n"
        b"--changeset_8b9e487e-a353-4dcb-a6f4-0688191e0314\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 2\r\n"
        b"\r\n"
        b"DELETE /container2/blob2 HTTP/1.1\r\n"
        b"\r\n"
        b"\r\n"
        b"--changeset_8b9e487e-a353-4dcb-a6f4-0688191e0314\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 3\r\n"
        b"\r\n"
        b"DELETE /container3/blob3 HTTP/1.1\r\n"
        b"\r\n"
        b"\r\n"
        b"--changeset_8b9e487e-a353-4dcb-a6f4-0688191e0314--\r\n"
        b"\r\n"
        b"--batch_357de4f7-6d0b-4e02-8cd2-6361411a9525--\r\n"
    )


@pytest.mark.asyncio
@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
async def test_multipart_send_with_combination_changeset_first(http_request):
    transport = MockAsyncHttpTransport()

    changeset = http_request("", "")
    changeset.set_multipart_mixed(
        http_request("DELETE", "/container0/blob0"),
        http_request("DELETE", "/container1/blob1"),
        boundary="changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525",
    )
    request = http_request("POST", "http://account.blob.core.windows.net/?comp=batch")
    request.set_multipart_mixed(
        changeset, http_request("DELETE", "/container2/blob2"), boundary="batch_357de4f7-6d0b-4e02-8cd2-6361411a9525"
    )

    async with AsyncPipeline(transport) as pipeline:
        await pipeline.run(request)

    assert request.body == (
        b"--batch_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: multipart/mixed; boundary=changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"\r\n"
        b"--changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 0\r\n"
        b"\r\n"
        b"DELETE /container0/blob0 HTTP/1.1\r\n"
        b"\r\n"
        b"\r\n"
        b"--changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 1\r\n"
        b"\r\n"
        b"DELETE /container1/blob1 HTTP/1.1\r\n"
        b"\r\n"
        b"\r\n"
        b"--changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525--\r\n"
        b"\r\n"
        b"--batch_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 2\r\n"
        b"\r\n"
        b"DELETE /container2/blob2 HTTP/1.1\r\n"
        b"\r\n"
        b"\r\n"
        b"--batch_357de4f7-6d0b-4e02-8cd2-6361411a9525--\r\n"
    )


@pytest.mark.asyncio
@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
async def test_multipart_send_with_combination_changeset_last(http_request):
    transport = MockAsyncHttpTransport()
    changeset = http_request("", "")
    changeset.set_multipart_mixed(
        http_request("DELETE", "/container1/blob1"),
        http_request("DELETE", "/container2/blob2"),
        boundary="changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525",
    )
    request = http_request("POST", "http://account.blob.core.windows.net/?comp=batch")
    request.set_multipart_mixed(
        http_request("DELETE", "/container0/blob0"), changeset, boundary="batch_357de4f7-6d0b-4e02-8cd2-6361411a9525"
    )

    async with AsyncPipeline(transport) as pipeline:
        await pipeline.run(request)

    assert request.body == (
        b"--batch_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 0\r\n"
        b"\r\n"
        b"DELETE /container0/blob0 HTTP/1.1\r\n"
        b"\r\n"
        b"\r\n"
        b"--batch_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: multipart/mixed; boundary=changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"\r\n"
        b"--changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 1\r\n"
        b"\r\n"
        b"DELETE /container1/blob1 HTTP/1.1\r\n"
        b"\r\n"
        b"\r\n"
        b"--changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 2\r\n"
        b"\r\n"
        b"DELETE /container2/blob2 HTTP/1.1\r\n"
        b"\r\n"
        b"\r\n"
        b"--changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525--\r\n"
        b"\r\n"
        b"--batch_357de4f7-6d0b-4e02-8cd2-6361411a9525--\r\n"
    )


@pytest.mark.asyncio
@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
async def test_multipart_send_with_combination_changeset_middle(http_request):
    transport = MockAsyncHttpTransport()
    changeset = http_request("", "")
    changeset.set_multipart_mixed(
        http_request("DELETE", "/container1/blob1"), boundary="changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525"
    )
    request = http_request("POST", "http://account.blob.core.windows.net/?comp=batch")
    request.set_multipart_mixed(
        http_request("DELETE", "/container0/blob0"),
        changeset,
        http_request("DELETE", "/container2/blob2"),
        boundary="batch_357de4f7-6d0b-4e02-8cd2-6361411a9525",
    )

    async with AsyncPipeline(transport) as pipeline:
        await pipeline.run(request)

    assert request.body == (
        b"--batch_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 0\r\n"
        b"\r\n"
        b"DELETE /container0/blob0 HTTP/1.1\r\n"
        b"\r\n"
        b"\r\n"
        b"--batch_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: multipart/mixed; boundary=changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"\r\n"
        b"--changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 1\r\n"
        b"\r\n"
        b"DELETE /container1/blob1 HTTP/1.1\r\n"
        b"\r\n"
        b"\r\n"
        b"--changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525--\r\n"
        b"\r\n"
        b"--batch_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 2\r\n"
        b"\r\n"
        b"DELETE /container2/blob2 HTTP/1.1\r\n"
        b"\r\n"
        b"\r\n"
        b"--batch_357de4f7-6d0b-4e02-8cd2-6361411a9525--\r\n"
    )


@pytest.mark.asyncio
@pytest.mark.parametrize("http_request,mock_response", request_and_responses_product(MOCK_RESPONSES))
async def test_multipart_receive(http_request, mock_response):
    class ResponsePolicy(object):
        def on_response(self, request, response):
            # type: (PipelineRequest, PipelineResponse) -> None
            response.http_response.headers["x-ms-fun"] = "true"

    class AsyncResponsePolicy(object):
        async def on_response(self, request, response):
            # type: (PipelineRequest, PipelineResponse) -> None
            response.http_response.headers["x-ms-async-fun"] = "true"

    req0 = http_request("DELETE", "/container0/blob0")
    req1 = http_request("DELETE", "/container1/blob1")

    request = http_request("POST", "http://account.blob.core.windows.net/?comp=batch")
    request.set_multipart_mixed(req0, req1, policies=[ResponsePolicy(), AsyncResponsePolicy()])

    body_as_str = (
        "--batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed\r\n"
        "Content-Type: application/http\r\n"
        "Content-ID: 0\r\n"
        "\r\n"
        "HTTP/1.1 202 Accepted\r\n"
        "x-ms-request-id: 778fdc83-801e-0000-62ff-0334671e284f\r\n"
        "x-ms-version: 2018-11-09\r\n"
        "\r\n"
        "--batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed\r\n"
        "Content-Type: application/http\r\n"
        "Content-ID: 2\r\n"
        "\r\n"
        "HTTP/1.1 404 The specified blob does not exist.\r\n"
        "x-ms-error-code: BlobNotFound\r\n"
        "x-ms-request-id: 778fdc83-801e-0000-62ff-0334671e2852\r\n"
        "x-ms-version: 2018-11-09\r\n"
        "Content-Length: 216\r\n"
        "Content-Type: application/xml\r\n"
        "\r\n"
        '<?xml version="1.0" encoding="utf-8"?>\r\n'
        "<Error><Code>BlobNotFound</Code><Message>The specified blob does not exist.\r\n"
        "RequestId:778fdc83-801e-0000-62ff-0334671e2852\r\n"
        "Time:2018-06-14T16:46:54.6040685Z</Message></Error>\r\n"
        "--batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed--"
    )

    response = mock_response(
        request,
        body_as_str.encode("ascii"),
        "multipart/mixed; boundary=batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed",
    )

    parts = []
    async for part in response.parts():
        parts.append(part)

    assert len(parts) == 2

    res0 = parts[0]
    assert res0.status_code == 202
    assert res0.headers["x-ms-fun"] == "true"
    assert res0.headers["x-ms-async-fun"] == "true"

    res1 = parts[1]
    assert res1.status_code == 404
    assert res1.headers["x-ms-fun"] == "true"
    assert res1.headers["x-ms-async-fun"] == "true"


@pytest.mark.asyncio
@pytest.mark.parametrize("http_request,mock_response", request_and_responses_product(MOCK_RESPONSES))
async def test_multipart_receive_with_one_changeset(http_request, mock_response):
    changeset = http_request("", "")
    changeset.set_multipart_mixed(
        http_request("DELETE", "/container0/blob0"), http_request("DELETE", "/container1/blob1")
    )

    request = http_request("POST", "http://account.blob.core.windows.net/?comp=batch")
    request.set_multipart_mixed(changeset)
    body_as_bytes = (
        b"--batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed\r\n"
        b'Content-Type: multipart/mixed; boundary="changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525"\r\n'
        b"\r\n"
        b"--changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 0\r\n"
        b"\r\n"
        b"HTTP/1.1 202 Accepted\r\n"
        b"x-ms-request-id: 778fdc83-801e-0000-62ff-0334671e284f\r\n"
        b"x-ms-version: 2018-11-09\r\n"
        b"\r\n"
        b"\r\n"
        b"--changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 1\r\n"
        b"\r\n"
        b"HTTP/1.1 202 Accepted\r\n"
        b"x-ms-request-id: 778fdc83-801e-0000-62ff-0334671e284f\r\n"
        b"x-ms-version: 2018-11-09\r\n"
        b"\r\n"
        b"\r\n"
        b"--changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525--\r\n"
        b"\r\n"
        b"--batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed--\r\n"
    )

    response = mock_response(
        request, body_as_bytes, "multipart/mixed; boundary=batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed"
    )

    parts = []
    async for part in response.parts():
        parts.append(part)
    assert len(parts) == 2

    res0 = parts[0]
    assert res0.status_code == 202


@pytest.mark.asyncio
@pytest.mark.parametrize("http_request,mock_response", request_and_responses_product(MOCK_RESPONSES))
async def test_multipart_receive_with_multiple_changesets(http_request, mock_response):
    changeset1 = http_request("", "")
    changeset1.set_multipart_mixed(
        http_request("DELETE", "/container0/blob0"), http_request("DELETE", "/container1/blob1")
    )
    changeset2 = http_request("", "")
    changeset2.set_multipart_mixed(
        http_request("DELETE", "/container2/blob2"), http_request("DELETE", "/container3/blob3")
    )

    request = http_request("POST", "http://account.blob.core.windows.net/?comp=batch")
    request.set_multipart_mixed(changeset1, changeset2)
    body_as_bytes = (
        b"--batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed\r\n"
        b'Content-Type: multipart/mixed; boundary="changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525"\r\n'
        b"\r\n"
        b"--changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 0\r\n"
        b"\r\n"
        b"HTTP/1.1 200\r\n"
        b"x-ms-request-id: 778fdc83-801e-0000-62ff-0334671e284f\r\n"
        b"x-ms-version: 2018-11-09\r\n"
        b"\r\n"
        b"\r\n"
        b"--changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 1\r\n"
        b"\r\n"
        b"HTTP/1.1 202\r\n"
        b"x-ms-request-id: 778fdc83-801e-0000-62ff-0334671e284f\r\n"
        b"x-ms-version: 2018-11-09\r\n"
        b"\r\n"
        b"\r\n"
        b"--changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525--\r\n"
        b"\r\n"
        b"--batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed\r\n"
        b'Content-Type: multipart/mixed; boundary="changeset_8b9e487e-a353-4dcb-a6f4-0688191e0314"\r\n'
        b"\r\n"
        b"--changeset_8b9e487e-a353-4dcb-a6f4-0688191e0314\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 2\r\n"
        b"\r\n"
        b"HTTP/1.1 404\r\n"
        b"x-ms-request-id: 778fdc83-801e-0000-62ff-0334671e284f\r\n"
        b"x-ms-version: 2018-11-09\r\n"
        b"\r\n"
        b"\r\n"
        b"--changeset_8b9e487e-a353-4dcb-a6f4-0688191e0314\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 3\r\n"
        b"\r\n"
        b"HTTP/1.1 409\r\n"
        b"x-ms-request-id: 778fdc83-801e-0000-62ff-0334671e284f\r\n"
        b"x-ms-version: 2018-11-09\r\n"
        b"\r\n"
        b"\r\n"
        b"--changeset_8b9e487e-a353-4dcb-a6f4-0688191e0314--\r\n"
        b"\r\n"
        b"--batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed--\r\n"
    )

    response = mock_response(
        request, body_as_bytes, "multipart/mixed; boundary=batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed"
    )

    parts = []
    async for part in response.parts():
        parts.append(part)
    assert len(parts) == 4
    assert parts[0].status_code == 200
    assert parts[1].status_code == 202
    assert parts[2].status_code == 404
    assert parts[3].status_code == 409


@pytest.mark.asyncio
@pytest.mark.parametrize("http_request,mock_response", request_and_responses_product(MOCK_RESPONSES))
async def test_multipart_receive_with_combination_changeset_first(http_request, mock_response):
    changeset = http_request("", "")
    changeset.set_multipart_mixed(
        http_request("DELETE", "/container0/blob0"), http_request("DELETE", "/container1/blob1")
    )

    request = http_request("POST", "http://account.blob.core.windows.net/?comp=batch")
    request.set_multipart_mixed(changeset, http_request("DELETE", "/container2/blob2"))
    body_as_bytes = (
        b"--batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed\r\n"
        b'Content-Type: multipart/mixed; boundary="changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525"\r\n'
        b"\r\n"
        b"--changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 0\r\n"
        b"\r\n"
        b"HTTP/1.1 200\r\n"
        b"x-ms-request-id: 778fdc83-801e-0000-62ff-0334671e284f\r\n"
        b"x-ms-version: 2018-11-09\r\n"
        b"\r\n"
        b"\r\n"
        b"--changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 1\r\n"
        b"\r\n"
        b"HTTP/1.1 202\r\n"
        b"x-ms-request-id: 778fdc83-801e-0000-62ff-0334671e284f\r\n"
        b"x-ms-version: 2018-11-09\r\n"
        b"\r\n"
        b"\r\n"
        b"--changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525--\r\n"
        b"\r\n"
        b"--batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 2\r\n"
        b"\r\n"
        b"HTTP/1.1 404\r\n"
        b"x-ms-request-id: 778fdc83-801e-0000-62ff-0334671e284f\r\n"
        b"x-ms-version: 2018-11-09\r\n"
        b"\r\n"
        b"\r\n"
        b"--batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed--\r\n"
    )

    response = mock_response(
        request, body_as_bytes, "multipart/mixed; boundary=batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed"
    )

    parts = []
    async for part in response.parts():
        parts.append(part)
    assert len(parts) == 3
    assert parts[0].status_code == 200
    assert parts[1].status_code == 202
    assert parts[2].status_code == 404


@pytest.mark.parametrize("mock_response", MOCK_RESPONSES)
def test_raise_for_status_bad_response(mock_response):
    response = mock_response(request=None, body=None, content_type=None)
    response.status_code = 400
    with pytest.raises(HttpResponseError):
        response.raise_for_status()


@pytest.mark.parametrize("mock_response", MOCK_RESPONSES)
def test_raise_for_status_good_response(mock_response):
    response = mock_response(request=None, body=None, content_type=None)
    response.status_code = 200
    response.raise_for_status()


@pytest.mark.asyncio
@pytest.mark.parametrize("http_request,mock_response", request_and_responses_product(MOCK_RESPONSES))
async def test_multipart_receive_with_combination_changeset_middle(http_request, mock_response):
    changeset = http_request("", "")
    changeset.set_multipart_mixed(http_request("DELETE", "/container1/blob1"))

    request = http_request("POST", "http://account.blob.core.windows.net/?comp=batch")
    request.set_multipart_mixed(
        http_request("DELETE", "/container0/blob0"), changeset, http_request("DELETE", "/container2/blob2")
    )
    body_as_bytes = (
        b"--batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 2\r\n"
        b"\r\n"
        b"HTTP/1.1 200\r\n"
        b"x-ms-request-id: 778fdc83-801e-0000-62ff-0334671e284f\r\n"
        b"x-ms-version: 2018-11-09\r\n"
        b"\r\n"
        b"\r\n"
        b"--batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed\r\n"
        b'Content-Type: multipart/mixed; boundary="changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525"\r\n'
        b"\r\n"
        b"--changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 0\r\n"
        b"\r\n"
        b"HTTP/1.1 202\r\n"
        b"x-ms-request-id: 778fdc83-801e-0000-62ff-0334671e284f\r\n"
        b"x-ms-version: 2018-11-09\r\n"
        b"\r\n"
        b"\r\n"
        b"--changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525--\r\n"
        b"\r\n"
        b"--batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 2\r\n"
        b"\r\n"
        b"HTTP/1.1 404\r\n"
        b"x-ms-request-id: 778fdc83-801e-0000-62ff-0334671e284f\r\n"
        b"x-ms-version: 2018-11-09\r\n"
        b"\r\n"
        b"\r\n"
        b"--batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed--\r\n"
    )

    response = mock_response(
        request, body_as_bytes, "multipart/mixed; boundary=batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed"
    )

    parts = []
    async for part in response.parts():
        parts.append(part)
    assert len(parts) == 3
    assert parts[0].status_code == 200
    assert parts[1].status_code == 202
    assert parts[2].status_code == 404


@pytest.mark.asyncio
@pytest.mark.parametrize("http_request,mock_response", request_and_responses_product(MOCK_RESPONSES))
async def test_multipart_receive_with_combination_changeset_last(http_request, mock_response):
    changeset = http_request("", "")
    changeset.set_multipart_mixed(
        http_request("DELETE", "/container1/blob1"), http_request("DELETE", "/container2/blob2")
    )

    request = http_request("POST", "http://account.blob.core.windows.net/?comp=batch")
    request.set_multipart_mixed(http_request("DELETE", "/container0/blob0"), changeset)

    body_as_bytes = (
        b"--batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 2\r\n"
        b"\r\n"
        b"HTTP/1.1 200\r\n"
        b"x-ms-request-id: 778fdc83-801e-0000-62ff-0334671e284f\r\n"
        b"x-ms-version: 2018-11-09\r\n"
        b"\r\n"
        b"\r\n"
        b"--batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed\r\n"
        b'Content-Type: multipart/mixed; boundary="changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525"\r\n'
        b"\r\n"
        b"--changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 0\r\n"
        b"\r\n"
        b"HTTP/1.1 202\r\n"
        b"x-ms-request-id: 778fdc83-801e-0000-62ff-0334671e284f\r\n"
        b"x-ms-version: 2018-11-09\r\n"
        b"\r\n"
        b"\r\n"
        b"--changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 1\r\n"
        b"\r\n"
        b"HTTP/1.1 404\r\n"
        b"x-ms-request-id: 778fdc83-801e-0000-62ff-0334671e284f\r\n"
        b"x-ms-version: 2018-11-09\r\n"
        b"\r\n"
        b"\r\n"
        b"--changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525--\r\n"
        b"\r\n"
        b"--batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed--\r\n"
    )

    response = mock_response(
        request, body_as_bytes, "multipart/mixed; boundary=batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed"
    )

    parts = []
    async for part in response.parts():
        parts.append(part)
    assert len(parts) == 3
    assert parts[0].status_code == 200
    assert parts[1].status_code == 202
    assert parts[2].status_code == 404


@pytest.mark.asyncio
@pytest.mark.parametrize("http_request,mock_response", request_and_responses_product(MOCK_RESPONSES))
async def test_multipart_receive_with_bom(http_request, mock_response):
    req0 = http_request("DELETE", "/container0/blob0")

    request = http_request("POST", "http://account.blob.core.windows.net/?comp=batch")
    request.set_multipart_mixed(req0)
    body_as_bytes = (
        b"--batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed\n"
        b"Content-Type: application/http\n"
        b"Content-Transfer-Encoding: binary\n"
        b"Content-ID: 0\n"
        b"\r\n"
        b"HTTP/1.1 400 One of the request inputs is not valid.\r\n"
        b"Content-Length: 220\r\n"
        b"Content-Type: application/xml\r\n"
        b"Server: Windows-Azure-Blob/1.0\r\n"
        b"\r\n"
        b'\xef\xbb\xbf<?xml version="1.0" encoding="utf-8"?>\n<Error><Code>InvalidInput</Code><Message>One'
        b"of the request inputs is not valid.\nRequestId:5f3f9f2f-e01e-00cc-6eb1-6d00b5000000\nTime:2019-09-17T23:44:07.4671860Z</Message></Error>\n"
        b"--batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed--"
    )

    response = mock_response(
        request, body_as_bytes, "multipart/mixed; boundary=batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed"
    )

    parts = []
    async for part in response.parts():
        parts.append(part)
    assert len(parts) == 1

    res0 = parts[0]
    assert res0.status_code == 400
    assert res0.body().startswith(b"\xef\xbb\xbf")


@pytest.mark.asyncio
@pytest.mark.parametrize("http_request,mock_response", request_and_responses_product(MOCK_RESPONSES))
async def test_recursive_multipart_receive(http_request, mock_response):
    req0 = http_request("DELETE", "/container0/blob0")
    internal_req0 = http_request("DELETE", "/container0/blob0")
    req0.set_multipart_mixed(internal_req0)

    request = http_request("POST", "http://account.blob.core.windows.net/?comp=batch")
    request.set_multipart_mixed(req0)
    internal_body_as_str = (
        "--batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed\r\n"
        "Content-Type: application/http\r\n"
        "Content-ID: 0\r\n"
        "\r\n"
        "HTTP/1.1 400 Accepted\r\n"
        "x-ms-request-id: 778fdc83-801e-0000-62ff-0334671e284f\r\n"
        "x-ms-version: 2018-11-09\r\n"
        "\r\n"
        "--batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed--"
    )

    body_as_str = (
        "--batchresponse_8d5f5bcd-2cb5-44bb-91b5-e9a722e68cb6\r\n"
        "Content-Type: application/http\r\n"
        "Content-ID: 0\r\n"
        "\r\n"
        "HTTP/1.1 202 Accepted\r\n"
        "Content-Type: multipart/mixed; boundary=batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed\r\n"
        "\r\n"
        "{}"
        "--batchresponse_8d5f5bcd-2cb5-44bb-91b5-e9a722e68cb6--"
    ).format(internal_body_as_str)

    response = mock_response(
        request,
        body_as_str.encode("ascii"),
        "multipart/mixed; boundary=batchresponse_8d5f5bcd-2cb5-44bb-91b5-e9a722e68cb6",
    )

    parts = []
    async for part in response.parts():
        parts.append(part)

    assert len(parts) == 1

    res0 = parts[0]
    assert res0.status_code == 202

    internal_parts = []
    async for part in res0.parts():
        internal_parts.append(part)
    assert len(internal_parts) == 1

    internal_response0 = internal_parts[0]
    assert internal_response0.status_code == 400


@pytest.mark.skipif(sys.version_info < (3, 10), reason="Loop parameter is deprecated since Python 3.10")
def test_aiohttp_loop():
    import asyncio
    from azure.core.pipeline.transport import AioHttpTransport

    loop = asyncio.get_event_loop()
    with pytest.raises(ValueError):
        transport = AioHttpTransport(loop=loop)


class MockAiohttpResponse:
    def __init__(self):
        self.status = 200
        self.reason = "OK"
        self.headers = {"content-type": "application/json"}
        self.content = MockAiohttpContent()

    def read(self):
        request_info = aiohttp.RequestInfo("http://example.org", "GET", {}, "http://example.org")
        raise aiohttp.client_exceptions.ClientResponseError(request_info, None)


class MockAiohttpContent:
    async def read(self, block_size):
        request_info = aiohttp.RequestInfo("http://example.org", "GET", {}, "http://example.org")
        raise aiohttp.client_exceptions.ClientResponseError(request_info, None)


@pytest.mark.asyncio
async def test_aiohttp_errors():
    request = HttpRequest("GET", "http://example.org")
    response = AioHttpTransportResponse(request, MockAiohttpResponse())
    with pytest.raises(ServiceResponseError):
        await response.load_body()

    generator = AioHttpStreamDownloadGenerator(None, response)
    with pytest.raises(ServiceResponseError):
        await generator.__anext__()



import pytest
from azure.core.utils import parse_connection_string


class TestCoreConnectionStringParser:
    # cSpell:disable
    def test_parsing_with_case_sensitive_keys_for_sensitive_conn_str(self, **kwargs):
        conn_str = "Endpoint=XXXXENDPOINTXXXX;SharedAccessKeyName=XXXXPOLICYXXXX;SharedAccessKey=THISISATESTKEYXXXXXXXXXXXXXXXXXXXXXXXXXXXX="
        parse_result = parse_connection_string(conn_str, True)
        assert parse_result["Endpoint"] == "XXXXENDPOINTXXXX"
        assert parse_result["SharedAccessKeyName"] == "XXXXPOLICYXXXX"
        assert parse_result["SharedAccessKey"] == "THISISATESTKEYXXXXXXXXXXXXXXXXXXXXXXXXXXXX="
        with pytest.raises(KeyError):
            parse_result["endPoint"]
        with pytest.raises(KeyError):
            parse_result["sharedAccESSkEynAME"]
        with pytest.raises(KeyError):
            parse_result["sharedaccesskey"]

    def test_parsing_with_case_insensitive_keys_for_sensitive_conn_str(self, **kwargs):
        conn_str = "Endpoint=XXXXENDPOINTXXXX;SharedAccessKeyName=XXXXPOLICYXXXX;SharedAccessKey=THISISATESTKEYXXXXXXXXXXXXXXXXXXXXXXXXXXXX="
        parse_result = parse_connection_string(conn_str, False)
        assert parse_result["endpoint"] == "XXXXENDPOINTXXXX"
        assert parse_result["sharedaccesskeyname"] == "XXXXPOLICYXXXX"
        assert parse_result["sharedaccesskey"] == "THISISATESTKEYXXXXXXXXXXXXXXXXXXXXXXXXXXXX="

    def test_parsing_with_case_insensitive_keys_for_insensitive_conn_str(self, **kwargs):
        conn_str = "enDpoiNT=XXXXENDPOINTXXXX;sharedaccesskeyname=XXXXPOLICYXXXX;SHAREDACCESSKEY=THISISATESTKEYXXXXXXXXXXXXXXXXXXXXXXXXXXXX="
        parse_result = parse_connection_string(conn_str, False)
        assert parse_result["endpoint"] == "XXXXENDPOINTXXXX"
        assert parse_result["sharedaccesskeyname"] == "XXXXPOLICYXXXX"
        assert parse_result["sharedaccesskey"] == "THISISATESTKEYXXXXXXXXXXXXXXXXXXXXXXXXXXXX="

    def test_error_with_duplicate_case_sensitive_keys_for_sensitive_conn_str(self, **kwargs):
        conn_str = "Endpoint=XXXXENDPOINTXXXX;Endpoint=XXXXENDPOINT2XXXX;SharedAccessKeyName=XXXXPOLICYXXXX;SharedAccessKey=THISISATESTKEYXXXXXXXXXXXXXXXXXXXXXXXXXXXX="
        with pytest.raises(ValueError) as e:
            parse_result = parse_connection_string(conn_str, True)
        assert str(e.value) == "Connection string is either blank or malformed."

    def test_success_with_duplicate_case_sensitive_keys_for_sensitive_conn_str(self, **kwargs):
        conn_str = "enDpoInt=XXXXENDPOINTXXXX;Endpoint=XXXXENDPOINT2XXXX;"
        parse_result = parse_connection_string(conn_str, True)
        assert parse_result["enDpoInt"] == "XXXXENDPOINTXXXX"
        assert parse_result["Endpoint"] == "XXXXENDPOINT2XXXX"

    def test_error_with_duplicate_case_insensitive_keys_for_insensitive_conn_str(self, **kwargs):
        conn_str = "endPoinT=XXXXENDPOINTXXXX;eNdpOint=XXXXENDPOINT2XXXX;sharedaccesskeyname=XXXXPOLICYXXXX;SHAREDACCESSKEY=THISISATESTKEYXXXXXXXXXXXXXXXXXXXXXXXXXXXX="
        with pytest.raises(ValueError) as e:
            parse_result = parse_connection_string(conn_str, False)
        assert str(e.value) == "Duplicate key in connection string: endpoint"

    def test_error_with_malformed_conn_str(self):
        for conn_str in ["", "foobar", "foo;bar;baz", ";", "foo=;bar=;", "=", "=;=="]:
            with pytest.raises(ValueError) as e:
                parse_result = parse_connection_string(conn_str)
            assert str(e.value) == "Connection string is either blank or malformed."

    def test_case_insensitive_clear_method(self):
        conn_str = "enDpoiNT=XXXXENDPOINTXXXX;sharedaccesskeyname=XXXXPOLICYXXXX;SHAREDACCESSKEY=THISISATESTKEYXXXXXXXXXXXXXXXXXXXXXXXXXXXX="
        parse_result = parse_connection_string(conn_str, False)
        parse_result.clear()
        assert len(parse_result) == 0

    def test_case_insensitive_copy_method(self):
        conn_str = "enDpoiNT=XXXXENDPOINTXXXX;sharedaccesskeyname=XXXXPOLICYXXXX;SHAREDACCESSKEY=THISISATESTKEYXXXXXXXXXXXXXXXXXXXXXXXXXXXX="
        parse_result = parse_connection_string(conn_str, False)
        copied = parse_result.copy()
        assert copied == parse_result

    def test_case_insensitive_get_method(self):
        conn_str = "Endpoint=XXXXENDPOINTXXXX;SharedAccessKeyName=XXXXPOLICYXXXX;SharedAccessKey=THISISATESTKEYXXXXXXXXXXXXXXXXXXXXXXXXXXXX="
        parse_result = parse_connection_string(conn_str, False)
        assert parse_result.get("sharedaccesskeyname") == "XXXXPOLICYXXXX"
        assert parse_result.get("sharedaccesskey") == "THISISATESTKEYXXXXXXXXXXXXXXXXXXXXXXXXXXXX="
        assert parse_result.get("accesskey") is None
        assert parse_result.get("accesskey", "XXothertestkeyXX=") == "XXothertestkeyXX="

    def test_case_insensitive_keys_method(self):
        conn_str = "enDpoiNT=XXXXENDPOINTXXXX;sharedaccesskeyname=XXXXPOLICYXXXX;SHAREDACCESSKEY=THISISATESTKEYXXXXXXXXXXXXXXXXXXXXXXXXXXXX="
        parse_result = parse_connection_string(conn_str, False)
        keys = parse_result.keys()
        assert len(keys) == 3
        assert "endpoint" in keys

    def test_case_insensitive_pop_method(self):
        conn_str = "enDpoiNT=XXXXENDPOINTXXXX;sharedaccesskeyname=XXXXPOLICYXXXX;SHAREDACCESSKEY=THISISATESTKEYXXXXXXXXXXXXXXXXXXXXXXXXXXXX="
        parse_result = parse_connection_string(conn_str, False)
        endpoint = parse_result.pop("endpoint")
        sharedaccesskey = parse_result.pop("sharedaccesskey")
        assert len(parse_result) == 1
        assert endpoint == "XXXXENDPOINTXXXX"
        assert sharedaccesskey == "THISISATESTKEYXXXXXXXXXXXXXXXXXXXXXXXXXXXX="

    def test_case_insensitive_update_with_insensitive_method(self):
        conn_str = "enDpoiNT=XXXXENDPOINTXXXX;sharedaccesskeyname=XXXXPOLICYXXXX;SHAREDACCESSKEY=THISISATESTKEYXXXXXXXXXXXXXXXXXXXXXXXXXXXX="
        conn_str2 = "hostName=XXXXENDPOINTXXXX;ACCessKEy=THISISATESTKEYXXXXXXXXXXXXXXXXXXXXXXXXXXXX=;"
        parse_result_insensitive = parse_connection_string(conn_str, False)
        parse_result_insensitive2 = parse_connection_string(conn_str2, False)

        parse_result_insensitive.update(parse_result_insensitive2)
        assert len(parse_result_insensitive) == 5
        assert parse_result_insensitive["hostname"] == "XXXXENDPOINTXXXX"
        assert parse_result_insensitive["accesskey"] == "THISISATESTKEYXXXXXXXXXXXXXXXXXXXXXXXXXXXX="

        # check that update replace duplicate case insensitive keys
        conn_str_duplicate_key = "endpoint=XXXXENDPOINT2XXXX;ACCessKEy=TestKey"
        parse_result_insensitive_dupe = parse_connection_string(conn_str_duplicate_key, False)
        parse_result_insensitive.update(parse_result_insensitive_dupe)
        assert parse_result_insensitive_dupe["endpoint"] == "XXXXENDPOINT2XXXX"
        assert parse_result_insensitive_dupe["accesskey"] == "TestKey"
        assert len(parse_result_insensitive) == 5

    def test_case_sensitive_update_with_insensitive_method(self):
        conn_str = "enDpoiNT=XXXXENDPOINTXXXX;sharedaccesskeyname=XXXXPOLICYXXXX;SHAREDACCESSKEY=THISISATESTKEYXXXXXXXXXXXXXXXXXXXXXXXXXXXX="
        conn_str2 = "hostName=XXXXENDPOINTXXXX;ACCessKEy=THISISATESTKEYXXXXXXXXXXXXXXXXXXXXXXXXXXXX=;"
        parse_result_insensitive = parse_connection_string(conn_str, False)
        parse_result_sensitive = parse_connection_string(conn_str2, True)

        parse_result_sensitive.update(parse_result_insensitive)
        assert len(parse_result_sensitive) == 5
        assert parse_result_sensitive["hostName"] == "XXXXENDPOINTXXXX"
        with pytest.raises(KeyError):
            parse_result_sensitive["hostname"]

    def test_case_insensitive_values_method(self):
        conn_str = "enDpoiNT=XXXXENDPOINTXXXX;sharedaccesskeyname=XXXXPOLICYXXXX;SHAREDACCESSKEY=THISISATESTKEYXXXXXXXXXXXXXXXXXXXXXXXXXXXX="
        parse_result = parse_connection_string(conn_str, False)
        values = parse_result.values()
        assert len(values) == 3

    # cSpell:enable



# -*- coding: utf-8 -*-
# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE.txt in the project root for
# license information.
# -------------------------------------------------------------------------

from flask import (
    Response,
    Blueprint,
    request,
)
from .helpers import jsonify, get_dict

basic_api = Blueprint("basic_api", __name__)


@basic_api.route("/string", methods=["GET"])
def string():
    return Response("Hello, world!", status=200, mimetype="text/plain")


@basic_api.route("/lines", methods=["GET"])
def lines():
    return Response("Hello,\nworld!", status=200, mimetype="text/plain")


@basic_api.route("/bytes", methods=["GET"])
def bytes():
    return Response("Hello, world!".encode(), status=200, mimetype="text/plain")


@basic_api.route("/html", methods=["GET"])
def html():
    return Response("<html><body>Hello, world!</html></body>", status=200, mimetype="text/html")


@basic_api.route("/json", methods=["GET"])
def json():
    return Response('{"greeting": "hello", "recipient": "world"}', status=200, mimetype="application/json")


@basic_api.route("/complicated-json", methods=["POST"])
def complicated_json():
    # thanks to Sean Kane for this test!
    assert request.json["EmptyByte"] == ""
    assert request.json["EmptyUnicode"] == ""
    assert request.json["SpacesOnlyByte"] == "   "
    assert request.json["SpacesOnlyUnicode"] == "   "
    assert request.json["SpacesBeforeByte"] == "   Text"
    assert request.json["SpacesBeforeUnicode"] == "   Text"
    assert request.json["SpacesAfterByte"] == "Text   "
    assert request.json["SpacesAfterUnicode"] == "Text   "
    assert request.json["SpacesBeforeAndAfterByte"] == "   Text   "
    assert request.json["SpacesBeforeAndAfterUnicode"] == "   Text   "
    assert request.json["啊齄丂狛"] == "ꀕ"
    assert request.json["RowKey"] == "test2"
    assert request.json["啊齄丂狛狜"] == "hello"
    assert request.json["singlequote"] == "a''''b"
    assert request.json["doublequote"] == 'a""""b'
    assert request.json["None"] == None

    return Response(status=200)


@basic_api.route("/headers", methods=["GET"])
def headers():
    return Response(
        status=200,
        headers={
            "lowercase-header": "lowercase",
            "ALLCAPS-HEADER": "ALLCAPS",
            "CamelCase-Header": "camelCase",
        },
    )


@basic_api.route("/anything", methods=["GET", "POST", "PUT", "DELETE", "PATCH", "TRACE"])
def anything():
    return jsonify(
        get_dict(
            "url",
            "args",
            "headers",
            "origin",
            "method",
            "form",
            "data",
            "files",
            "json",
        )
    )



# ------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# ------------------------------------
"""The tests for decorators_async.py"""

try:
    from unittest import mock
except ImportError:
    import mock

import sys
import time

import pytest
from azure.core.pipeline import Pipeline, PipelineResponse
from azure.core.pipeline.policies import HTTPPolicy
from azure.core.pipeline.transport import HttpTransport
from azure.core.settings import settings
from azure.core.tracing import SpanKind
from azure.core.tracing.decorator import distributed_trace
from azure.core.tracing.decorator_async import distributed_trace_async
from tracing_common import FakeSpan
from utils import HTTP_REQUESTS


@pytest.fixture(scope="module")
def fake_span():
    settings.tracing_implementation.set_value(FakeSpan)


class MockClient:
    @distributed_trace
    def __init__(self, http_request, policies=None, assert_current_span=False):
        time.sleep(0.001)
        self.request = http_request("GET", "http://localhost")
        if policies is None:
            policies = []
        policies.append(mock.Mock(spec=HTTPPolicy, send=self.verify_request))
        self.policies = policies
        self.transport = mock.Mock(spec=HttpTransport)
        self.pipeline = Pipeline(self.transport, policies=policies)

        self.expected_response = mock.Mock(spec=PipelineResponse)
        self.assert_current_span = assert_current_span

    def verify_request(self, request):
        if self.assert_current_span:
            assert execution_context.get_current_span() is not None
        return self.expected_response

    @distributed_trace_async
    async def make_request(self, numb_times, **kwargs):
        time.sleep(0.001)
        if numb_times < 1:
            return None
        response = self.pipeline.run(self.request, **kwargs)
        await self.get_foo(merge_span=True)
        kwargs["merge_span"] = True
        await self.make_request(numb_times - 1, **kwargs)
        return response

    @distributed_trace_async
    async def merge_span_method(self):
        return await self.get_foo(merge_span=True)

    @distributed_trace_async
    async def no_merge_span_method(self):
        return await self.get_foo()

    @distributed_trace_async
    async def get_foo(self):
        time.sleep(0.001)
        return 5

    @distributed_trace_async(name_of_span="different name")
    async def check_name_is_different(self):
        time.sleep(0.001)

    @distributed_trace_async(tracing_attributes={"foo": "bar"})
    async def tracing_attr(self):
        time.sleep(0.001)

    @distributed_trace_async(kind=SpanKind.PRODUCER)
    async def kind_override(self):
        time.sleep(0.001)

    @distributed_trace_async
    async def raising_exception(self):
        raise ValueError("Something went horribly wrong here")


@pytest.mark.usefixtures("fake_span")
class TestAsyncDecorator(object):
    @pytest.mark.asyncio
    @pytest.mark.parametrize("http_request", HTTP_REQUESTS)
    async def test_decorator_tracing_attr(self, http_request):
        with FakeSpan(name="parent") as parent:
            client = MockClient(http_request)
            await client.tracing_attr()

        assert len(parent.children) == 2
        assert parent.children[0].name == "MockClient.__init__"
        assert parent.children[1].name == "MockClient.tracing_attr"
        assert parent.children[1].kind == SpanKind.INTERNAL
        assert parent.children[1].attributes == {"foo": "bar"}

    @pytest.mark.asyncio
    @pytest.mark.parametrize("http_request", HTTP_REQUESTS)
    async def test_decorator_has_different_name(self, http_request):
        with FakeSpan(name="parent") as parent:
            client = MockClient(http_request)
            await client.check_name_is_different()
        assert len(parent.children) == 2
        assert parent.children[0].name == "MockClient.__init__"
        assert parent.children[1].name == "different name"
        assert parent.children[1].kind == SpanKind.INTERNAL

    @pytest.mark.asyncio
    @pytest.mark.parametrize("http_request", HTTP_REQUESTS)
    async def test_kind_override(self, http_request):
        with FakeSpan(name="parent") as parent:
            client = MockClient(http_request)
            await client.kind_override()

        assert len(parent.children) == 2
        assert parent.children[0].name == "MockClient.__init__"
        assert parent.children[1].name == "MockClient.kind_override"
        assert parent.children[1].kind == SpanKind.PRODUCER

    @pytest.mark.asyncio
    @pytest.mark.parametrize("http_request", HTTP_REQUESTS)
    async def test_used(self, http_request):
        with FakeSpan(name="parent") as parent:
            client = MockClient(http_request, policies=[])
            await client.get_foo(parent_span=parent)
            await client.get_foo()

            assert len(parent.children) == 3
            assert parent.children[0].name == "MockClient.__init__"
            assert not parent.children[0].children
            assert parent.children[1].name == "MockClient.get_foo"
            assert not parent.children[1].children
            assert parent.children[2].name == "MockClient.get_foo"
            assert not parent.children[2].children

    @pytest.mark.asyncio
    @pytest.mark.parametrize("http_request", HTTP_REQUESTS)
    async def test_span_merge_span(self, http_request):
        with FakeSpan(name="parent") as parent:
            client = MockClient(http_request)
            await client.merge_span_method()
            await client.no_merge_span_method()

        assert len(parent.children) == 3
        assert parent.children[0].name == "MockClient.__init__"
        assert not parent.children[0].children
        assert parent.children[1].name == "MockClient.merge_span_method"
        assert not parent.children[1].children
        assert parent.children[2].name == "MockClient.no_merge_span_method"
        assert parent.children[2].children[0].name == "MockClient.get_foo"

    @pytest.mark.asyncio
    @pytest.mark.parametrize("http_request", HTTP_REQUESTS)
    async def test_span_complicated(self, http_request):
        with FakeSpan(name="parent") as parent:
            client = MockClient(http_request)
            await client.make_request(2)
            with parent.span("child") as child:
                time.sleep(0.001)
                await client.make_request(2, parent_span=parent)
                assert FakeSpan.get_current_span() == child
                await client.make_request(2)

        assert len(parent.children) == 4
        assert parent.children[0].name == "MockClient.__init__"
        assert not parent.children[0].children
        assert parent.children[1].name == "MockClient.make_request"
        assert not parent.children[1].children
        assert parent.children[2].name == "child"
        assert parent.children[2].children[0].name == "MockClient.make_request"
        assert parent.children[3].name == "MockClient.make_request"
        assert not parent.children[3].children

    @pytest.mark.asyncio
    @pytest.mark.parametrize("http_request", HTTP_REQUESTS)
    async def test_span_with_exception(self, http_request):
        """Assert that if an exception is raised, the next sibling method is actually a sibling span."""
        with FakeSpan(name="parent") as parent:
            client = MockClient(http_request)
            try:
                await client.raising_exception()
            except:
                pass
            await client.get_foo()

        assert len(parent.children) == 3
        assert parent.children[0].name == "MockClient.__init__"
        assert parent.children[1].name == "MockClient.raising_exception"
        # Exception should propagate status for Opencensus
        assert parent.children[1].status == "Something went horribly wrong here"
        assert parent.children[2].name == "MockClient.get_foo"



# ------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# ------------------------------------
"""Tests for the distributed tracing policy."""
import logging

from azure.core.pipeline import PipelineResponse, PipelineRequest, PipelineContext
from azure.core.pipeline.policies import DistributedTracingPolicy, UserAgentPolicy
from azure.core.settings import settings
from tracing_common import FakeSpan
import time
import pytest
from utils import HTTP_RESPONSES, create_http_response, request_and_responses_product

try:
    from unittest import mock
except ImportError:
    import mock


@pytest.mark.parametrize("http_request,http_response", request_and_responses_product(HTTP_RESPONSES))
def test_distributed_tracing_policy_solo(http_request, http_response):
    """Test policy with no other policy and happy path"""
    settings.tracing_implementation.set_value(FakeSpan)
    with FakeSpan(name="parent") as root_span:
        policy = DistributedTracingPolicy()

        request = http_request("GET", "http://localhost/temp?query=query")
        request.headers["x-ms-client-request-id"] = "some client request id"

        pipeline_request = PipelineRequest(request, PipelineContext(None))
        policy.on_request(pipeline_request)

        response = create_http_response(http_response, request, None)
        response.headers = request.headers
        response.status_code = 202
        response.headers["x-ms-request-id"] = "some request id"

        assert request.headers.get("traceparent") == "123456789"

        policy.on_response(pipeline_request, PipelineResponse(request, response, PipelineContext(None)))
        time.sleep(0.001)
        policy.on_request(pipeline_request)
        try:
            raise ValueError("Transport trouble")
        except:
            policy.on_exception(pipeline_request)

    # Check on_response
    network_span = root_span.children[0]
    assert network_span.name == "/temp"
    assert network_span.attributes.get("http.method") == "GET"
    assert network_span.attributes.get("component") == "http"
    assert network_span.attributes.get("http.url") == "http://localhost/temp?query=query"
    assert network_span.attributes.get("http.user_agent") is None
    assert network_span.attributes.get("x-ms-request-id") == "some request id"
    assert network_span.attributes.get("x-ms-client-request-id") == "some client request id"
    assert network_span.attributes.get("http.status_code") == 202

    # Check on_exception
    network_span = root_span.children[1]
    assert network_span.name == "/temp"
    assert network_span.attributes.get("http.method") == "GET"
    assert network_span.attributes.get("component") == "http"
    assert network_span.attributes.get("http.url") == "http://localhost/temp?query=query"
    assert network_span.attributes.get("x-ms-client-request-id") == "some client request id"
    assert network_span.attributes.get("http.user_agent") is None
    assert network_span.attributes.get("x-ms-request-id") == None
    assert network_span.attributes.get("http.status_code") == 504


@pytest.mark.parametrize("http_request,http_response", request_and_responses_product(HTTP_RESPONSES))
def test_distributed_tracing_policy_attributes(http_request, http_response):
    """Test policy with no other policy and happy path"""
    settings.tracing_implementation.set_value(FakeSpan)
    with FakeSpan(name="parent") as root_span:
        policy = DistributedTracingPolicy(tracing_attributes={"myattr": "myvalue"})

        request = http_request("GET", "http://localhost/temp?query=query")

        pipeline_request = PipelineRequest(request, PipelineContext(None))
        policy.on_request(pipeline_request)

        response = create_http_response(http_response, request, None)
        response.headers = request.headers
        response.status_code = 202

        policy.on_response(pipeline_request, PipelineResponse(request, response, PipelineContext(None)))

    # Check on_response
    network_span = root_span.children[0]
    assert network_span.attributes.get("myattr") == "myvalue"


@pytest.mark.parametrize("http_request,http_response", request_and_responses_product(HTTP_RESPONSES))
def test_distributed_tracing_policy_badurl(caplog, http_request, http_response):
    """Test policy with a bad url that will throw, and be sure policy ignores it"""
    settings.tracing_implementation.set_value(FakeSpan)
    with FakeSpan(name="parent") as root_span:
        policy = DistributedTracingPolicy()

        request = http_request("GET", "http://[[[")
        request.headers["x-ms-client-request-id"] = "some client request id"

        pipeline_request = PipelineRequest(request, PipelineContext(None))
        with caplog.at_level(logging.WARNING, logger="azure.core.pipeline.policies.distributed_tracing"):
            policy.on_request(pipeline_request)
        assert "Unable to start network span" in caplog.text

        response = create_http_response(http_response, request, None)
        response.headers = request.headers
        response.status_code = 202
        response.headers["x-ms-request-id"] = "some request id"

        assert request.headers.get("traceparent") is None  # Got not network trace

        policy.on_response(pipeline_request, PipelineResponse(request, response, PipelineContext(None)))
        time.sleep(0.001)

        policy.on_request(pipeline_request)
        try:
            raise ValueError("Transport trouble")
        except:
            policy.on_exception(pipeline_request)

    assert len(root_span.children) == 0


@pytest.mark.parametrize("http_request,http_response", request_and_responses_product(HTTP_RESPONSES))
def test_distributed_tracing_policy_with_user_agent(http_request, http_response):
    """Test policy working with user agent."""
    settings.tracing_implementation.set_value(FakeSpan)
    with mock.patch.dict("os.environ", {"AZURE_HTTP_USER_AGENT": "mytools"}):
        with FakeSpan(name="parent") as root_span:
            policy = DistributedTracingPolicy()

            request = http_request("GET", "http://localhost")
            request.headers["x-ms-client-request-id"] = "some client request id"

            pipeline_request = PipelineRequest(request, PipelineContext(None))

            user_agent = UserAgentPolicy()
            user_agent.on_request(pipeline_request)
            policy.on_request(pipeline_request)

            response = create_http_response(http_response, request, None)
            response.headers = request.headers
            response.status_code = 202
            response.headers["x-ms-request-id"] = "some request id"
            pipeline_response = PipelineResponse(request, response, PipelineContext(None))

            assert request.headers.get("traceparent") == "123456789"

            policy.on_response(pipeline_request, pipeline_response)

            time.sleep(0.001)
            policy.on_request(pipeline_request)
            try:
                raise ValueError("Transport trouble")
            except:
                policy.on_exception(pipeline_request)

            user_agent.on_response(pipeline_request, pipeline_response)

        network_span = root_span.children[0]
        assert network_span.name == "/"
        assert network_span.attributes.get("http.method") == "GET"
        assert network_span.attributes.get("component") == "http"
        assert network_span.attributes.get("http.url") == "http://localhost"
        assert network_span.attributes.get("http.user_agent").endswith("mytools")
        assert network_span.attributes.get("x-ms-request-id") == "some request id"
        assert network_span.attributes.get("x-ms-client-request-id") == "some client request id"
        assert network_span.attributes.get("http.status_code") == 202

        network_span = root_span.children[1]
        assert network_span.name == "/"
        assert network_span.attributes.get("http.method") == "GET"
        assert network_span.attributes.get("component") == "http"
        assert network_span.attributes.get("http.url") == "http://localhost"
        assert network_span.attributes.get("http.user_agent").endswith("mytools")
        assert network_span.attributes.get("x-ms-client-request-id") == "some client request id"
        assert network_span.attributes.get("x-ms-request-id") is None
        assert network_span.attributes.get("http.status_code") == 504
        # Exception should propagate status for Opencensus
        assert network_span.status == "Transport trouble"


@pytest.mark.parametrize("http_request,http_response", request_and_responses_product(HTTP_RESPONSES))
def test_span_namer(http_request, http_response):
    settings.tracing_implementation.set_value(FakeSpan)
    with FakeSpan(name="parent") as root_span:

        request = http_request("GET", "http://localhost/temp?query=query")
        pipeline_request = PipelineRequest(request, PipelineContext(None))

        def fixed_namer(http_request):
            assert http_request is request
            return "overriddenname"

        policy = DistributedTracingPolicy(network_span_namer=fixed_namer)

        policy.on_request(pipeline_request)

        response = create_http_response(http_response, request, None)
        response.headers = request.headers
        response.status_code = 202

        policy.on_response(pipeline_request, PipelineResponse(request, response, PipelineContext(None)))

        def operation_namer(http_request):
            assert http_request is request
            return "operation level name"

        pipeline_request.context.options["network_span_namer"] = operation_namer

        policy.on_request(pipeline_request)

        response = create_http_response(http_response, request, None)
        response.headers = request.headers
        response.status_code = 202

        policy.on_response(pipeline_request, PipelineResponse(request, response, PipelineContext(None)))

    # Check init kwargs
    network_span = root_span.children[0]
    assert network_span.name == "overriddenname"

    # Check operation kwargs
    network_span = root_span.children[1]
    assert network_span.name == "operation level name"



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE.txt in the project root for
# license information.
# -------------------------------------------------------------------------
from azure.core.pipeline.transport import AsyncioRequestsTransport
from azure.core.rest import HttpRequest
from azure.core.rest._requests_asyncio import RestAsyncioRequestsTransportResponse
from rest_client_async import AsyncMockRestClient

import pytest
import pytest_asyncio
from utils import readonly_checks


@pytest_asyncio.fixture
async def client(port):
    async with AsyncioRequestsTransport() as transport:
        async with AsyncMockRestClient(port, transport=transport) as client:
            yield client


@pytest.mark.asyncio
async def test_async_gen_data(port, client):
    class AsyncGen:
        def __init__(self):
            self._range = iter([b"azerty"])

        def __aiter__(self):
            return self

        async def __anext__(self):
            try:
                return next(self._range)
            except StopIteration:
                raise StopAsyncIteration

    request = HttpRequest("GET", "http://localhost:{}/basic/anything".format(port), content=AsyncGen())
    response = await client.send_request(request)
    assert response.json()["data"] == "azerty"


@pytest.mark.asyncio
async def test_send_data(port, client):
    request = HttpRequest("PUT", "http://localhost:{}/basic/anything".format(port), content=b"azerty")
    response = await client.send_request(request)

    assert response.json()["data"] == "azerty"


@pytest.mark.asyncio
async def test_readonly(client):
    """Make sure everything that is readonly is readonly"""
    response = await client.send_request(HttpRequest("GET", "/health"))
    response.raise_for_status()

    assert isinstance(response, RestAsyncioRequestsTransportResponse)
    from azure.core.pipeline.transport import AsyncioRequestsTransportResponse

    readonly_checks(response, old_response_class=AsyncioRequestsTransportResponse)


@pytest.mark.asyncio
async def test_decompress_compressed_header(client):
    # expect plain text
    request = HttpRequest("GET", "/encoding/gzip")
    response = await client.send_request(request)
    content = await response.read()
    assert content == b"hello world"
    assert response.content == content
    assert response.text() == "hello world"


@pytest.mark.asyncio
async def test_deflate_decompress_compressed_header(client):
    # expect plain text
    request = HttpRequest("GET", "/encoding/deflate")
    response = await client.send_request(request)
    content = await response.read()
    assert content == b"hi there"
    assert response.content == content
    assert response.text() == "hi there"


@pytest.mark.asyncio
async def test_decompress_compressed_header_stream(client):
    # expect plain text
    request = HttpRequest("GET", "/encoding/gzip")
    response = await client.send_request(request, stream=True)
    content = await response.read()
    assert content == b"hello world"
    assert response.content == content
    assert response.text() == "hello world"


@pytest.mark.asyncio
async def test_decompress_compressed_header_stream_body_content(client):
    # expect plain text
    request = HttpRequest("GET", "/encoding/gzip")
    response = await client.send_request(request, stream=True)
    await response.read()
    content = response.content
    assert content == response.body()



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE.txt in the project root for
# license information.
# -------------------------------------------------------------------------

# NOTE: These tests are heavily inspired from the httpx test suite: https://github.com/encode/httpx/tree/master/tests
# Thank you httpx for your wonderful tests!
import io
import pytest
import zlib
from azure.core.rest import HttpRequest, AsyncHttpResponse
from azure.core.rest._aiohttp import RestAioHttpTransportResponse
from azure.core.exceptions import HttpResponseError
from utils import readonly_checks


@pytest.fixture
def send_request(client):
    async def _send_request(request):
        async with client:
            response = await client.send_request(request, stream=False)
            response.raise_for_status()
            return response

    return _send_request


@pytest.mark.asyncio
async def test_response(send_request, port):
    response = await send_request(
        HttpRequest("GET", "/basic/string"),
    )
    assert response.status_code == 200
    assert response.reason == "OK"
    assert response.content == b"Hello, world!"
    assert response.text() == "Hello, world!"
    assert response.request.method == "GET"
    assert response.request.url == "http://localhost:{}/basic/string".format(port)


@pytest.mark.asyncio
async def test_response_content(send_request):
    response = await send_request(
        request=HttpRequest("GET", "/basic/bytes"),
    )
    assert response.status_code == 200
    assert response.reason == "OK"
    content = await response.read()
    assert content == b"Hello, world!"
    assert response.text() == "Hello, world!"


@pytest.mark.asyncio
async def test_response_text(send_request):
    response = await send_request(
        request=HttpRequest("GET", "/basic/string"),
    )
    assert response.status_code == 200
    assert response.reason == "OK"
    content = await response.read()
    assert content == b"Hello, world!"
    assert response.text() == "Hello, world!"
    assert response.headers["Content-Length"] == "13"
    assert response.headers["Content-Type"] == "text/plain; charset=utf-8"


@pytest.mark.asyncio
async def test_response_html(send_request):
    response = await send_request(
        request=HttpRequest("GET", "/basic/html"),
    )
    assert response.status_code == 200
    assert response.reason == "OK"
    content = await response.read()
    assert content == b"<html><body>Hello, world!</html></body>"
    assert response.text() == "<html><body>Hello, world!</html></body>"


@pytest.mark.asyncio
async def test_raise_for_status(client):
    # response = await client.send_request(
    #     HttpRequest("GET", "/basic/string"),
    # )
    # response.raise_for_status()

    response = await client.send_request(
        HttpRequest("GET", "/errors/403"),
    )
    assert response.status_code == 403
    with pytest.raises(HttpResponseError):
        response.raise_for_status()

    response = await client.send_request(
        HttpRequest("GET", "/errors/500"),
        retry_total=0,  # takes too long with retires on 500
    )
    assert response.status_code == 500
    with pytest.raises(HttpResponseError):
        response.raise_for_status()


@pytest.mark.asyncio
async def test_response_repr(send_request):
    response = await send_request(HttpRequest("GET", "/basic/string"))
    assert repr(response) == "<AsyncHttpResponse: 200 OK, Content-Type: text/plain; charset=utf-8>"


@pytest.mark.asyncio
async def test_response_content_type_encoding(send_request):
    """
    Use the charset encoding in the Content-Type header if possible.
    """
    response = await send_request(request=HttpRequest("GET", "/encoding/latin-1"))
    assert response.content_type == "text/plain; charset=latin-1"
    assert response.content == b"Latin 1: \xff"
    assert response.text() == "Latin 1: ÿ"
    assert response.encoding == "latin-1"


@pytest.mark.asyncio
async def test_response_autodetect_encoding(send_request):
    """
    Autodetect encoding if there is no Content-Type header.
    """
    response = await send_request(request=HttpRequest("GET", "/encoding/latin-1"))
    assert response.text() == "Latin 1: ÿ"
    assert response.encoding == "latin-1"


@pytest.mark.asyncio
async def test_response_fallback_to_autodetect(send_request):
    """
    Fallback to autodetection if we get an invalid charset in the Content-Type header.
    """
    response = await send_request(request=HttpRequest("GET", "/encoding/invalid-codec-name"))
    assert response.headers["Content-Type"] == "text/plain; charset=invalid-codec-name"
    assert response.text() == "おはようございます。"
    assert response.encoding is None


@pytest.mark.asyncio
async def test_response_no_charset_with_ascii_content(send_request):
    """
    A response with ascii encoded content should decode correctly,
    even with no charset specified.
    """
    response = await send_request(
        request=HttpRequest("GET", "/encoding/no-charset"),
    )

    assert response.headers["Content-Type"] == "text/plain"
    assert response.status_code == 200
    assert response.encoding is None
    content = await response.read()
    assert content == b"Hello, world!"
    assert response.text() == "Hello, world!"


@pytest.mark.asyncio
async def test_response_no_charset_with_iso_8859_1_content(send_request):
    """
    We don't support iso-8859-1 by default following conversations
    about encoding flow
    """
    response = await send_request(
        request=HttpRequest("GET", "/encoding/iso-8859-1"),
    )
    assert response.text() == "Accented: �sterreich"
    assert response.encoding is None


@pytest.mark.asyncio
async def test_json(send_request):
    response = await send_request(
        request=HttpRequest("GET", "/basic/json"),
    )
    assert response.json() == {"greeting": "hello", "recipient": "world"}
    assert response.encoding is None


@pytest.mark.asyncio
async def test_json_with_specified_encoding(send_request):
    response = await send_request(
        request=HttpRequest("GET", "/encoding/json"),
    )
    assert response.json() == {"greeting": "hello", "recipient": "world"}
    assert response.encoding == "utf-16"


@pytest.mark.asyncio
async def test_emoji(send_request):
    response = await send_request(
        request=HttpRequest("GET", "/encoding/emoji"),
    )
    assert response.text() == "👩"


@pytest.mark.asyncio
async def test_emoji_family_with_skin_tone_modifier(send_request):
    response = await send_request(
        request=HttpRequest("GET", "/encoding/emoji-family-skin-tone-modifier"),
    )
    assert response.text() == "👩🏻‍👩🏽‍👧🏾‍👦🏿 SSN: 859-98-0987"


@pytest.mark.asyncio
async def test_korean_nfc(send_request):
    response = await send_request(
        request=HttpRequest("GET", "/encoding/korean"),
    )
    assert response.text() == "아가"


@pytest.mark.asyncio
async def test_urlencoded_content(send_request):
    await send_request(
        request=HttpRequest(
            "POST", "/urlencoded/pet/add/1", data={"pet_type": "dog", "pet_food": "meat", "name": "Fido", "pet_age": 42}
        ),
    )


@pytest.mark.asyncio
async def test_multipart_files_content(send_request):
    request = HttpRequest(
        "POST",
        "/multipart/basic",
        files={"fileContent": io.BytesIO(b"<file content>")},
    )
    await send_request(request)


@pytest.mark.asyncio
async def test_multipart_data_and_files_content(send_request):
    request = HttpRequest(
        "POST",
        "/multipart/data-and-files",
        data={"message": "Hello, world!"},
        files={"fileContent": io.BytesIO(b"<file content>")},
    )
    await send_request(request)


@pytest.mark.asyncio
async def test_send_request_return_pipeline_response(client):
    # we use return_pipeline_response for some cases in autorest
    request = HttpRequest("GET", "/basic/string")
    response = await client.send_request(request, _return_pipeline_response=True)
    assert hasattr(response, "http_request")
    assert hasattr(response, "http_response")
    assert hasattr(response, "context")
    assert response.http_response.text() == "Hello, world!"
    assert hasattr(response.http_request, "content")


@pytest.mark.asyncio
async def test_text_and_encoding(send_request):
    response = await send_request(
        request=HttpRequest("GET", "/encoding/emoji"),
    )
    assert response.content == "👩".encode("utf-8")
    assert response.text() == "👩"

    # try setting encoding as a property
    response.encoding = "utf-16"
    assert response.text() == "鿰ꦑ" == response.content.decode(response.encoding)

    # assert latin-1 changes text decoding without changing encoding property
    assert response.text("latin-1") == "ð\x9f\x91©" == response.content.decode("latin-1")
    assert response.encoding == "utf-16"


# @pytest.mark.asyncio
# async def test_multipart_encode_non_seekable_filelike(send_request):
#     """
#     Test that special readable but non-seekable filelike objects are supported,
#     at the cost of reading them into memory at most once.
#     """

#     class IteratorIO(io.IOBase):
#         def __init__(self, iterator):
#             self._iterator = iterator

#         def read(self, *args):
#             return b"".join(self._iterator)

#     def data():
#         yield b"Hello"
#         yield b"World"

#     fileobj = IteratorIO(data())
#     files = {"file": fileobj}
#     request = HttpRequest(
#         "POST",
#         "/multipart/non-seekable-filelike",
#         files=files,
#     )
#     await send_request(request)


def test_initialize_response_abc():
    with pytest.raises(TypeError) as ex:
        AsyncHttpResponse()
    assert "Can't instantiate abstract class" in str(ex)


@pytest.mark.asyncio
async def test_readonly(send_request):
    """Make sure everything that is readonly is readonly"""
    response = await send_request(HttpRequest("GET", "/health"))

    assert isinstance(response, RestAioHttpTransportResponse)
    from azure.core.pipeline.transport import AioHttpTransportResponse

    readonly_checks(response, old_response_class=AioHttpTransportResponse)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE.txt in the project root for
# license information.
# -------------------------------------------------------------------------
from azure.core.exceptions import HttpResponseError, ServiceRequestError
import pytest
from azure.core.rest import HttpRequest
from azure.core.exceptions import StreamClosedError, StreamConsumedError, ResponseNotReadError


@pytest.mark.asyncio
async def test_iter_raw(client):
    request = HttpRequest("GET", "/streams/basic")
    async with client.send_request(request, stream=True) as response:
        raw = b""
        async for part in response.iter_raw():
            raw += part
        assert raw == b"Hello, world!"


@pytest.mark.asyncio
async def test_iter_raw_on_iterable(client):
    request = HttpRequest("GET", "/streams/iterable")

    async with client.send_request(request, stream=True) as response:
        raw = b""
        async for part in response.iter_raw():
            raw += part
        assert raw == b"Hello, world!"


@pytest.mark.asyncio
async def test_iter_with_error(client):
    request = HttpRequest("GET", "/errors/403")

    async with client.send_request(request, stream=True) as response:
        try:
            response.raise_for_status()
        except HttpResponseError as e:
            pass
    assert response.is_closed

    try:
        async with client.send_request(request, stream=True) as response:
            response.raise_for_status()
    except HttpResponseError as e:
        pass

    assert response.is_closed

    request = HttpRequest("GET", "http://doesNotExist")
    with pytest.raises(ServiceRequestError):
        async with (await client.send_request(request, stream=True)):
            raise ValueError("Should error before entering")
    assert response.is_closed


@pytest.mark.asyncio
async def test_iter_bytes(client):
    request = HttpRequest("GET", "/streams/basic")

    async with client.send_request(request, stream=True) as response:
        raw = b""
        async for chunk in response.iter_bytes():
            assert response.is_stream_consumed
            assert not response.is_closed
            raw += chunk
        assert response.is_stream_consumed
        assert response.is_closed
        assert raw == b"Hello, world!"


@pytest.mark.skip(reason="We've gotten rid of iter_text for now")
@pytest.mark.asyncio
async def test_iter_text(client):
    request = HttpRequest("GET", "/basic/string")

    async with client.send_request(request, stream=True) as response:
        content = ""
        async for part in response.iter_text():
            content += part
        assert content == "Hello, world!"


@pytest.mark.skip(reason="We've gotten rid of iter_lines for now")
@pytest.mark.asyncio
async def test_iter_lines(client):
    request = HttpRequest("GET", "/basic/lines")

    async with client.send_request(request, stream=True) as response:
        content = []
        async for part in response.iter_lines():
            content.append(part)
        assert content == ["Hello,\n", "world!"]


@pytest.mark.asyncio
async def test_streaming_response(client):
    request = HttpRequest("GET", "/streams/basic")

    async with client.send_request(request, stream=True) as response:
        assert response.status_code == 200
        assert not response.is_closed

        content = await response.read()
        assert content == b"Hello, world!"
        assert response.content == b"Hello, world!"
        assert response.is_closed


@pytest.mark.asyncio
async def test_cannot_read_after_stream_consumed(port, client):
    request = HttpRequest("GET", "/streams/basic")
    async with client.send_request(request, stream=True) as response:
        content = b""
        async for chunk in response.iter_bytes():
            content += chunk

        with pytest.raises(StreamConsumedError) as ex:
            await response.read()
    assert "<HttpRequest [GET], url: 'http://localhost:{}/streams/basic'>".format(port) in str(ex.value)
    assert "You have likely already consumed this stream, so it can not be accessed anymore" in str(ex.value)


@pytest.mark.asyncio
async def test_cannot_read_after_response_closed(port, client):
    request = HttpRequest("GET", "/streams/basic")
    async with client.send_request(request, stream=True) as response:
        pass

    with pytest.raises(StreamClosedError) as ex:
        await response.read()
    assert "<HttpRequest [GET], url: 'http://localhost:{}/streams/basic'>".format(port) in str(ex.value)
    assert "can no longer be read or streamed, since the response has already been closed" in str(ex.value)


@pytest.mark.asyncio
async def test_decompress_plain_no_header(client):
    # thanks to Xiang Yan for this test!
    account_name = "coretests"
    url = "https://{}.blob.core.windows.net/tests/test.txt".format(account_name)
    request = HttpRequest("GET", url)
    async with client:
        response = await client.send_request(request, stream=True)
        with pytest.raises(ResponseNotReadError):
            response.content
        await response.read()
        assert response.content == b"test"


@pytest.mark.asyncio
async def test_compress_plain_no_header(client):
    # thanks to Xiang Yan for this test!
    account_name = "coretests"
    url = "https://{}.blob.core.windows.net/tests/test.txt".format(account_name)
    request = HttpRequest("GET", url)
    async with client:
        response = await client.send_request(request, stream=True)
        iter = response.iter_raw()
        data = b""
        async for d in iter:
            data += d
        assert data == b"test"


@pytest.mark.asyncio
async def test_iter_read_back_and_forth(client):
    # thanks to McCoy Patiño for this test!

    # while this test may look like it's exposing buggy behavior, this is httpx's behavior
    # the reason why the code flow is like this, is because the 'iter_x' functions don't
    # actually read the contents into the response, the output them. Once they're yielded,
    # the stream is closed, so you have to catch the output when you iterate through it
    request = HttpRequest("GET", "/basic/string")

    async with client.send_request(request, stream=True) as response:
        async for part in response.iter_bytes():
            assert part
        with pytest.raises(ResponseNotReadError):
            response.text()
        with pytest.raises(StreamConsumedError):
            await response.read()
        with pytest.raises(ResponseNotReadError):
            response.text()


@pytest.mark.asyncio
async def test_stream_with_return_pipeline_response(client):
    request = HttpRequest("GET", "/basic/string")
    pipeline_response = await client.send_request(request, stream=True, _return_pipeline_response=True)
    assert hasattr(pipeline_response, "http_request")
    assert hasattr(pipeline_response.http_request, "content")
    assert hasattr(pipeline_response, "http_response")
    assert hasattr(pipeline_response, "context")
    parts = []
    async for part in pipeline_response.http_response.iter_bytes():
        parts.append(part)
    assert parts == [b"Hello, world!"]
    await client.close()


@pytest.mark.asyncio
async def test_error_reading(client):
    request = HttpRequest("GET", "/errors/403")
    async with client.send_request(request, stream=True) as response:
        await response.read()
        assert response.content == b""
    response.content

    response = await client.send_request(request, stream=True)
    with pytest.raises(HttpResponseError):
        response.raise_for_status()
    await response.read()
    assert response.content == b""
    await client.close()


@pytest.mark.asyncio
async def test_pass_kwarg_to_iter_bytes(client):
    request = HttpRequest("GET", "/basic/string")
    response = await client.send_request(request, stream=True)
    async for part in response.iter_bytes(chunk_size=5):
        assert part


@pytest.mark.asyncio
async def test_pass_kwarg_to_iter_raw(client):
    request = HttpRequest("GET", "/basic/string")
    response = await client.send_request(request, stream=True)
    async for part in response.iter_raw(chunk_size=5):
        assert part


@pytest.mark.asyncio
async def test_decompress_compressed_header(client):
    # expect plain text
    request = HttpRequest("GET", "/encoding/gzip")
    response = await client.send_request(request)
    content = await response.read()
    assert content == b"hello world"
    assert response.content == content
    assert response.text() == "hello world"


@pytest.mark.asyncio
async def test_deflate_decompress_compressed_header(client):
    # expect plain text
    request = HttpRequest("GET", "/encoding/deflate")
    response = await client.send_request(request)
    content = await response.read()
    assert content == b"hi there"
    assert response.content == content
    assert response.text() == "hi there"


@pytest.mark.asyncio
async def test_decompress_compressed_header_stream(client):
    # expect plain text
    request = HttpRequest("GET", "/encoding/gzip")
    response = await client.send_request(request, stream=True)
    content = await response.read()
    assert content == b"hello world"
    assert response.content == content
    assert response.text() == "hello world"


@pytest.mark.asyncio
async def test_decompress_compressed_header_stream_body_content(client):
    # expect plain text
    request = HttpRequest("GET", "/encoding/gzip")
    response = await client.send_request(request, stream=True)
    await response.read()
    content = response.content
    assert content == response.body()



# coding: utf-8
# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE.txt in the project root for
# license information.
# -------------------------------------------------------------------------
from azure.core.pipeline import AsyncPipeline
from azure.core.pipeline.transport import (
    HttpRequest,
)
from azure.core import AsyncPipelineClient
from azure.core.exceptions import IncompleteReadError
import pytest


@pytest.mark.asyncio
async def test_aio_transport_short_read_download_stream(port):
    url = "http://localhost:{}/errors/short-data".format(port)
    client = AsyncPipelineClient(url)
    with pytest.raises(IncompleteReadError):
        async with client:
            request = HttpRequest("GET", url)
            pipeline_response = await client._pipeline.run(request, stream=True)
            response = pipeline_response.http_response
            data = response.stream_download(client._pipeline)
            content = b""
            async for d in data:
                content += d



# --------------------------------------------------------------------------
#
# Copyright (c) Microsoft Corporation. All rights reserved.
#
# The MIT License (MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the ""Software""), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# --------------------------------------------------------------------------
import pytest
from azure.core.exceptions import ServiceRequestError
from azure.core.rest import HttpRequest
from azure.core.polling import LROPoller
from azure.core.polling.base_polling import LROBasePolling


@pytest.fixture
def deserialization_callback():
    def _callback(response):
        return response.http_response.json()

    return _callback


@pytest.fixture
def lro_poller(client, deserialization_callback):
    def _callback(request, **kwargs):
        initial_response = client.send_request(request=request, _return_pipeline_response=True)
        return LROPoller(
            client._client,
            initial_response,
            deserialization_callback,
            LROBasePolling(0, **kwargs),
        )

    return _callback


def test_post_with_location_and_operation_location_headers(lro_poller):
    poller = lro_poller(HttpRequest("POST", "/polling/post/location-and-operation-location"))
    result = poller.result()
    assert result == {"location_result": True}


def test_post_with_location_and_operation_location_headers_no_body(lro_poller):
    poller = lro_poller(HttpRequest("POST", "/polling/post/location-and-operation-location-no-body"))
    result = poller.result()
    assert result is None


def test_post_resource_location(lro_poller):
    poller = lro_poller(HttpRequest("POST", "/polling/post/resource-location"))
    result = poller.result()
    assert result == {"location_result": True}


def test_put_no_polling(lro_poller):
    result = lro_poller(HttpRequest("PUT", "/polling/no-polling")).result()
    assert result["properties"]["provisioningState"] == "Succeeded"


def test_put_location(lro_poller):
    result = lro_poller(HttpRequest("PUT", "/polling/location")).result()
    assert result["location_result"]


def test_put_initial_response_body_invalid(lro_poller):
    # initial body is invalid
    result = lro_poller(HttpRequest("PUT", "/polling/initial-body-invalid")).result()
    assert result["location_result"]


def test_put_operation_location_polling_fail(lro_poller):
    with pytest.raises(ServiceRequestError):
        lro_poller(HttpRequest("PUT", "/polling/bad-operation-location"), retry_total=0).result()


def test_put_location_polling_fail(lro_poller):
    with pytest.raises(ServiceRequestError):
        lro_poller(HttpRequest("PUT", "/polling/bad-location"), retry_total=0).result()


def test_patch_location(lro_poller):
    result = lro_poller(HttpRequest("PATCH", "/polling/location")).result()
    assert result["location_result"]


def test_patch_operation_location_polling_fail(lro_poller):
    with pytest.raises(ServiceRequestError):
        lro_poller(HttpRequest("PUT", "/polling/bad-operation-location"), retry_total=0).result()


def test_patch_location_polling_fail(lro_poller):
    with pytest.raises(ServiceRequestError):
        lro_poller(HttpRequest("PUT", "/polling/bad-location"), retry_total=0).result()


def test_delete_operation_location(lro_poller):
    result = lro_poller(HttpRequest("DELETE", "/polling/operation-location")).result()
    assert result["status"] == "Succeeded"


def test_request_id(lro_poller):
    result = lro_poller(HttpRequest("POST", "/polling/request-id"), request_id="123456789").result()


def test_continuation_token(client, lro_poller, deserialization_callback):
    poller = lro_poller(HttpRequest("POST", "/polling/post/location-and-operation-location"))
    token = poller.continuation_token()
    new_poller = LROPoller.from_continuation_token(
        continuation_token=token,
        polling_method=LROBasePolling(0),
        client=client._client,
        deserialization_callback=deserialization_callback,
    )
    result = new_poller.result()
    assert result == {"location_result": True}



# --------------------------------------------------------------------------
#
# Copyright (c) Microsoft Corporation. All rights reserved.
#
# The MIT License (MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the ""Software""), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.
#
# --------------------------------------------------------------------------
import time
import pytest
import signal
import os
import subprocess
import sys
import random
import urllib
from rest_client_async import AsyncMockRestClient


def is_port_available(port_num):
    req = urllib.request.Request("http://localhost:{}/health".format(port_num))
    try:
        return urllib.request.urlopen(req).code != 200
    except Exception as e:
        return True


def get_port():
    count = 3
    for _ in range(count):
        port_num = random.randrange(3000, 5000)
        if is_port_available(port_num):
            return port_num
    raise TypeError("Tried {} times, can't find an open port".format(count))


@pytest.fixture
def port():
    return os.environ["FLASK_PORT"]


def start_testserver():
    port = get_port()
    os.environ["FLASK_APP"] = "coretestserver"
    os.environ["FLASK_PORT"] = str(port)
    cmd = "flask run -p {}".format(port)
    if os.name == "nt":  # On windows, subprocess creation works without being in the shell
        child_process = subprocess.Popen(cmd, env=dict(os.environ))
    else:
        # On linux, have to set shell=True
        child_process = subprocess.Popen(cmd, shell=True, preexec_fn=os.setsid, env=dict(os.environ))
    count = 5
    for _ in range(count):
        if not is_port_available(port):
            return child_process
        time.sleep(1)
    raise ValueError("Didn't start!")


def terminate_testserver(process):
    if os.name == "nt":
        process.kill()
    else:
        os.killpg(os.getpgid(process.pid), signal.SIGTERM)  # Send the signal to all the process groups


@pytest.fixture(autouse=True, scope="package")
def testserver():
    """Start the Autorest testserver."""
    server = start_testserver()
    yield
    terminate_testserver(server)


@pytest.fixture
def client(port):
    return AsyncMockRestClient(port)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE.txt in the project root for
# license information.
# -------------------------------------------------------------------------
import time
from itertools import product
from requests import Response
import azure.core
from azure.core.credentials import AccessToken, AzureKeyCredential, AzureSasCredential, AzureNamedKeyCredential
from azure.core.exceptions import ServiceRequestError
from azure.core.pipeline import Pipeline
from azure.core.pipeline.transport import HttpTransport, HttpRequest
from azure.core.pipeline.policies import (
    BearerTokenCredentialPolicy,
    RedirectPolicy,
    SansIOHTTPPolicy,
    AzureKeyCredentialPolicy,
    AzureSasCredentialPolicy,
    SensitiveHeaderCleanupPolicy,
)
from utils import HTTP_REQUESTS

import pytest

from unittest.mock import Mock


@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
def test_bearer_policy_adds_header(http_request):
    """The bearer token policy should add a header containing a token from its credential"""
    # 2524608000 == 01/01/2050 @ 12:00am (UTC)
    expected_token = AccessToken("expected_token", 2524608000)

    def verify_authorization_header(request):
        assert request.http_request.headers["Authorization"] == "Bearer {}".format(expected_token.token)
        return Mock()

    fake_credential = Mock(get_token=Mock(return_value=expected_token))
    policies = [BearerTokenCredentialPolicy(fake_credential, "scope"), Mock(send=verify_authorization_header)]

    pipeline = Pipeline(transport=Mock(), policies=policies)
    pipeline.run(http_request("GET", "https://spam.eggs"))

    assert fake_credential.get_token.call_count == 1

    pipeline.run(http_request("GET", "https://spam.eggs"))

    # Didn't need a new token
    assert fake_credential.get_token.call_count == 1


@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
def test_bearer_policy_send(http_request):
    """The bearer token policy should invoke the next policy's send method and return the result"""
    expected_request = http_request("GET", "https://spam.eggs")
    expected_response = Mock()

    def verify_request(request):
        assert request.http_request is expected_request
        return expected_response

    def get_token(*_, **__):
        return AccessToken("***", 42)

    fake_credential = Mock(get_token=get_token)
    policies = [BearerTokenCredentialPolicy(fake_credential, "scope"), Mock(send=verify_request)]
    response = Pipeline(transport=Mock(), policies=policies).run(expected_request)

    assert response is expected_response


@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
def test_bearer_policy_token_caching(http_request):
    good_for_one_hour = AccessToken("token", time.time() + 3600)
    credential = Mock(get_token=Mock(return_value=good_for_one_hour))
    pipeline = Pipeline(transport=Mock(), policies=[BearerTokenCredentialPolicy(credential, "scope")])

    pipeline.run(http_request("GET", "https://spam.eggs"))
    assert credential.get_token.call_count == 1  # policy has no token at first request -> it should call get_token

    pipeline.run(http_request("GET", "https://spam.eggs"))
    assert credential.get_token.call_count == 1  # token is good for an hour -> policy should return it from cache

    expired_token = AccessToken("token", time.time())
    credential.get_token.reset_mock()
    credential.get_token.return_value = expired_token
    pipeline = Pipeline(transport=Mock(), policies=[BearerTokenCredentialPolicy(credential, "scope")])

    pipeline.run(http_request("GET", "https://spam.eggs"))
    assert credential.get_token.call_count == 1

    pipeline.run(http_request("GET", "https://spam.eggs"))
    assert credential.get_token.call_count == 2  # token expired -> policy should call get_token


@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
def test_bearer_policy_optionally_enforces_https(http_request):
    """HTTPS enforcement should be controlled by a keyword argument, and enabled by default"""

    def assert_option_popped(request, **kwargs):
        assert "enforce_https" not in kwargs, "BearerTokenCredentialPolicy didn't pop the 'enforce_https' option"
        return Mock()

    def get_token(*_, **__):
        return AccessToken("***", 42)

    credential = Mock(get_token=get_token)
    pipeline = Pipeline(
        transport=Mock(send=assert_option_popped), policies=[BearerTokenCredentialPolicy(credential, "scope")]
    )

    # by default and when enforce_https=True, the policy should raise when given an insecure request
    with pytest.raises(ServiceRequestError):
        pipeline.run(http_request("GET", "http://not.secure"))
    with pytest.raises(ServiceRequestError):
        pipeline.run(http_request("GET", "http://not.secure"), enforce_https=True)

    # when enforce_https=False, an insecure request should pass
    pipeline.run(http_request("GET", "http://not.secure"), enforce_https=False)

    # https requests should always pass
    pipeline.run(http_request("GET", "https://secure"), enforce_https=False)
    pipeline.run(http_request("GET", "https://secure"), enforce_https=True)
    pipeline.run(http_request("GET", "https://secure"))


@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
def test_bearer_policy_preserves_enforce_https_opt_out(http_request):
    """The policy should use request context to preserve an opt out from https enforcement"""

    class ContextValidator(SansIOHTTPPolicy):
        def on_request(self, request):
            assert "enforce_https" in request.context, "'enforce_https' is not in the request's context"
            return Mock()

    credential = Mock(get_token=Mock(return_value=AccessToken("***", 42)))
    policies = [BearerTokenCredentialPolicy(credential, "scope"), ContextValidator()]
    pipeline = Pipeline(transport=Mock(), policies=policies)

    pipeline.run(http_request("GET", "http://not.secure"), enforce_https=False)


@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
def test_bearer_policy_default_context(http_request):
    """The policy should call get_token with the scopes given at construction, and no keyword arguments, by default"""
    expected_scope = "scope"
    token = AccessToken("", 0)
    credential = Mock(get_token=Mock(return_value=token))
    policy = BearerTokenCredentialPolicy(credential, expected_scope)
    pipeline = Pipeline(transport=Mock(), policies=[policy])

    pipeline.run(http_request("GET", "https://localhost"))

    credential.get_token.assert_called_once_with(expected_scope)


@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
def test_bearer_policy_enable_cae(http_request):
    """The policy should set enable_cae to True in the get_token request if it is set in constructor."""
    expected_scope = "scope"
    token = AccessToken("", 0)
    credential = Mock(get_token=Mock(return_value=token))
    policy = BearerTokenCredentialPolicy(credential, expected_scope, enable_cae=True)
    pipeline = Pipeline(transport=Mock(), policies=[policy])

    pipeline.run(http_request("GET", "https://localhost"))

    credential.get_token.assert_called_once_with(expected_scope, enable_cae=True)


@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
def test_bearer_policy_context_unmodified_by_default(http_request):
    """When no options for the policy accompany a request, the policy shouldn't add anything to the request context"""

    class ContextValidator(SansIOHTTPPolicy):
        def on_request(self, request):
            assert not any(request.context), "the policy shouldn't add to the request's context"

    credential = Mock(get_token=Mock(return_value=AccessToken("***", 42)))
    policies = [BearerTokenCredentialPolicy(credential, "scope"), ContextValidator()]
    pipeline = Pipeline(transport=Mock(), policies=policies)

    pipeline.run(http_request("GET", "https://secure"))


@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
def test_bearer_policy_calls_on_challenge(http_request):
    """BearerTokenCredentialPolicy should call its on_challenge method when it receives an authentication challenge"""

    class TestPolicy(BearerTokenCredentialPolicy):
        called = False

        def on_challenge(self, request, challenge):
            self.__class__.called = True
            return False

    credential = Mock(get_token=Mock(return_value=AccessToken("***", int(time.time()) + 3600)))
    policies = [TestPolicy(credential, "scope")]
    response = Mock(status_code=401, headers={"WWW-Authenticate": 'Basic realm="localhost"'})
    transport = Mock(send=Mock(return_value=response))

    pipeline = Pipeline(transport=transport, policies=policies)
    pipeline.run(http_request("GET", "https://localhost"))

    assert TestPolicy.called


@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
def test_bearer_policy_cannot_complete_challenge(http_request):
    """BearerTokenCredentialPolicy should return the 401 response when it can't complete its challenge"""

    expected_scope = "scope"
    expected_token = AccessToken("***", int(time.time()) + 3600)
    credential = Mock(get_token=Mock(return_value=expected_token))
    expected_response = Mock(status_code=401, headers={"WWW-Authenticate": 'Basic realm="localhost"'})
    transport = Mock(send=Mock(return_value=expected_response))
    policies = [BearerTokenCredentialPolicy(credential, expected_scope)]

    pipeline = Pipeline(transport=transport, policies=policies)
    response = pipeline.run(http_request("GET", "https://localhost"))

    assert response.http_response is expected_response
    assert transport.send.call_count == 1
    credential.get_token.assert_called_once_with(expected_scope)


@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
def test_bearer_policy_calls_sansio_methods(http_request):
    """BearerTokenCredentialPolicy should call SansIOHttpPolicy methods as does _SansIOHTTPPolicyRunner"""

    class TestPolicy(BearerTokenCredentialPolicy):
        def __init__(self, *args, **kwargs):
            super(TestPolicy, self).__init__(*args, **kwargs)
            self.on_exception = Mock(return_value=False)
            self.on_request = Mock()
            self.on_response = Mock()

        def send(self, request):
            self.request = request
            self.response = super(TestPolicy, self).send(request)
            return self.response

    credential = Mock(get_token=Mock(return_value=AccessToken("***", int(time.time()) + 3600)))
    policy = TestPolicy(credential, "scope")
    transport = Mock(send=Mock(return_value=Mock(status_code=200)))

    pipeline = Pipeline(transport=transport, policies=[policy])
    pipeline.run(http_request("GET", "https://localhost"))

    policy.on_request.assert_called_once_with(policy.request)
    policy.on_response.assert_called_once_with(policy.request, policy.response)

    # the policy should call on_exception when next.send() raises
    class TestException(Exception):
        pass

    # during the first send...
    transport = Mock(send=Mock(side_effect=TestException))
    policy = TestPolicy(credential, "scope")
    pipeline = Pipeline(transport=transport, policies=[policy])
    with pytest.raises(TestException):
        pipeline.run(http_request("GET", "https://localhost"))
    policy.on_exception.assert_called_once_with(policy.request)

    # ...or the second
    def raise_the_second_time(*args, **kwargs):
        if raise_the_second_time.calls == 0:
            raise_the_second_time.calls = 1
            return Mock(status_code=401, headers={"WWW-Authenticate": 'Basic realm="localhost"'})
        raise TestException()

    raise_the_second_time.calls = 0

    policy = TestPolicy(credential, "scope")
    policy.on_challenge = Mock(return_value=True)
    transport = Mock(send=Mock(wraps=raise_the_second_time))
    pipeline = Pipeline(transport=transport, policies=[policy])
    with pytest.raises(TestException):
        pipeline.run(http_request("GET", "https://localhost"))
    assert transport.send.call_count == 2
    policy.on_challenge.assert_called_once()
    policy.on_exception.assert_called_once_with(policy.request)


@pytest.mark.skipif(azure.core.__version__ >= "2", reason="this test applies only to azure-core 1.x")
@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
def test_key_vault_regression(http_request):
    """Test for regression affecting azure-keyvault-* 4.0.0. This test must pass, unmodified, for all 1.x versions."""

    from azure.core.pipeline.policies._authentication import _BearerTokenCredentialPolicyBase

    credential = Mock()
    policy = _BearerTokenCredentialPolicyBase(credential)
    assert policy._credential is credential

    headers = {}
    token = "alphanums"  # cspell:disable-line
    policy._update_headers(headers, token)
    assert headers["Authorization"] == "Bearer " + token

    assert policy._need_new_token
    policy._token = AccessToken(token, time.time() + 3600)
    assert not policy._need_new_token
    assert policy._token.token == token


@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
def test_azure_key_credential_policy(http_request):
    """Tests to see if we can create an AzureKeyCredentialPolicy"""

    key_header = "api_key"
    api_key = "test_key"

    def verify_authorization_header(request):
        assert request.headers[key_header] == api_key

    transport = Mock(send=verify_authorization_header)
    credential = AzureKeyCredential(api_key)
    credential_policy = AzureKeyCredentialPolicy(credential=credential, name=key_header)
    pipeline = Pipeline(transport=transport, policies=[credential_policy])

    pipeline.run(http_request("GET", "https://test_key_credential"))


def test_azure_key_credential_policy_raises():
    """Tests AzureKeyCredential and AzureKeyCredentialPolicy raises with non-compliant input parameters."""
    api_key = 1234
    key_header = 5678
    with pytest.raises(TypeError):
        credential = AzureKeyCredential(api_key)

    credential = AzureKeyCredential(str(api_key))
    with pytest.raises(TypeError):
        credential_policy = AzureKeyCredentialPolicy(credential=credential, name=key_header)

    with pytest.raises(TypeError):
        credential_policy = AzureKeyCredentialPolicy(credential=str(api_key), name=key_header)


def test_azure_key_credential_updates():
    """Tests AzureKeyCredential updates"""
    api_key = "original"

    credential = AzureKeyCredential(api_key)
    assert credential.key == api_key

    api_key = "new"
    credential.update(api_key)
    assert credential.key == api_key


combinations = [
    ("sig=test_signature", "https://test_sas_credential", "https://test_sas_credential?sig=test_signature"),
    ("?sig=test_signature", "https://test_sas_credential", "https://test_sas_credential?sig=test_signature"),
    (
        "sig=test_signature",
        "https://test_sas_credential?sig=test_signature",
        "https://test_sas_credential?sig=test_signature",
    ),
    (
        "?sig=test_signature",
        "https://test_sas_credential?sig=test_signature",
        "https://test_sas_credential?sig=test_signature",
    ),
    ("sig=test_signature", "https://test_sas_credential?", "https://test_sas_credential?sig=test_signature"),
    ("?sig=test_signature", "https://test_sas_credential?", "https://test_sas_credential?sig=test_signature"),
    (
        "sig=test_signature",
        "https://test_sas_credential?foo=bar",
        "https://test_sas_credential?foo=bar&sig=test_signature",
    ),
    (
        "?sig=test_signature",
        "https://test_sas_credential?foo=bar",
        "https://test_sas_credential?foo=bar&sig=test_signature",
    ),
]


@pytest.mark.parametrize("combinations,http_request", product(combinations, HTTP_REQUESTS))
def test_azure_sas_credential_policy(combinations, http_request):
    """Tests to see if we can create an AzureSasCredentialPolicy"""
    sas, url, expected_url = combinations

    def verify_authorization(request):
        assert request.url == expected_url

    transport = Mock(send=verify_authorization)
    credential = AzureSasCredential(sas)
    credential_policy = AzureSasCredentialPolicy(credential=credential)
    pipeline = Pipeline(transport=transport, policies=[credential_policy])

    pipeline.run(http_request("GET", url))


def test_azure_sas_credential_updates():
    """Tests AzureSasCredential updates"""
    sas = "original"

    credential = AzureSasCredential(sas)
    assert credential.signature == sas

    sas = "new"
    credential.update(sas)
    assert credential.signature == sas


def test_azure_sas_credential_policy_raises():
    """Tests AzureSasCredential and AzureSasCredentialPolicy raises with non-string input parameters."""
    sas = 1234
    with pytest.raises(TypeError):
        credential = AzureSasCredential(sas)


def test_azure_named_key_credential():
    cred = AzureNamedKeyCredential("sample_name", "samplekey")

    assert cred.named_key.name == "sample_name"
    assert cred.named_key.key == "samplekey"
    assert isinstance(cred.named_key, tuple)

    cred.update("newname", "newkey")
    assert cred.named_key.name == "newname"
    assert cred.named_key.key == "newkey"
    assert isinstance(cred.named_key, tuple)


def test_azure_named_key_credential_raises():
    with pytest.raises(TypeError, match="Both name and key must be strings."):
        cred = AzureNamedKeyCredential("sample_name", 123345)

    cred = AzureNamedKeyCredential("sample_name", "samplekey")
    assert cred.named_key.name == "sample_name"
    assert cred.named_key.key == "samplekey"

    with pytest.raises(TypeError, match="Both name and key must be strings."):
        cred.update(1234, "newkey")


def test_bearer_policy_redirect_same_domain():
    class MockTransport(HttpTransport):
        def __init__(self):
            self._first = True

        def __exit__(self, exc_type, exc_val, exc_tb):
            pass

        def close(self):
            pass

        def open(self):
            pass

        def send(self, request, **kwargs):  # type: (PipelineRequest, Any) -> PipelineResponse
            if self._first:
                self._first = False
                assert request.headers["Authorization"] == "Bearer {}".format(auth_headder)
                response = Response()
                response.status_code = 301
                response.headers["location"] = "https://localhost"
                return response
            assert request.headers["Authorization"] == "Bearer {}".format(auth_headder)
            response = Response()
            response.status_code = 200
            return response

    auth_headder = "token"
    expected_scope = "scope"
    token = AccessToken(auth_headder, 0)
    credential = Mock(get_token=Mock(return_value=token))
    auth_policy = BearerTokenCredentialPolicy(credential, expected_scope)
    redirect_policy = RedirectPolicy()
    header_clean_up_policy = SensitiveHeaderCleanupPolicy()
    pipeline = Pipeline(transport=MockTransport(), policies=[redirect_policy, auth_policy, header_clean_up_policy])

    pipeline.run(HttpRequest("GET", "https://localhost"))


def test_bearer_policy_redirect_different_domain():
    class MockTransport(HttpTransport):
        def __init__(self):
            self._first = True

        def __exit__(self, exc_type, exc_val, exc_tb):
            pass

        def close(self):
            pass

        def open(self):
            pass

        def send(self, request, **kwargs):  # type: (PipelineRequest, Any) -> PipelineResponse
            if self._first:
                self._first = False
                assert request.headers["Authorization"] == "Bearer {}".format(auth_headder)
                response = Response()
                response.status_code = 301
                response.headers["location"] = "https://localhost1"
                return response
            assert not request.headers.get("Authorization")
            response = Response()
            response.status_code = 200
            return response

    auth_headder = "token"
    expected_scope = "scope"
    token = AccessToken(auth_headder, 0)
    credential = Mock(get_token=Mock(return_value=token))
    auth_policy = BearerTokenCredentialPolicy(credential, expected_scope)
    redirect_policy = RedirectPolicy()
    header_clean_up_policy = SensitiveHeaderCleanupPolicy()
    pipeline = Pipeline(transport=MockTransport(), policies=[redirect_policy, auth_policy, header_clean_up_policy])

    pipeline.run(HttpRequest("GET", "https://localhost"))


def test_bearer_policy_redirect_opt_out_clean_up():
    class MockTransport(HttpTransport):
        def __init__(self):
            self._first = True

        def __exit__(self, exc_type, exc_val, exc_tb):
            pass

        def close(self):
            pass

        def open(self):
            pass

        def send(self, request, **kwargs):  # type: (PipelineRequest, Any) -> PipelineResponse
            if self._first:
                self._first = False
                assert request.headers["Authorization"] == "Bearer {}".format(auth_headder)
                response = Response()
                response.status_code = 301
                response.headers["location"] = "https://localhost1"
                return response
            assert request.headers["Authorization"] == "Bearer {}".format(auth_headder)
            response = Response()
            response.status_code = 200
            return response

    auth_headder = "token"
    expected_scope = "scope"
    token = AccessToken(auth_headder, 0)
    credential = Mock(get_token=Mock(return_value=token))
    auth_policy = BearerTokenCredentialPolicy(credential, expected_scope)
    redirect_policy = RedirectPolicy()
    header_clean_up_policy = SensitiveHeaderCleanupPolicy(disable_redirect_cleanup=True)
    pipeline = Pipeline(transport=MockTransport(), policies=[redirect_policy, auth_policy, header_clean_up_policy])

    pipeline.run(HttpRequest("GET", "https://localhost"))


def test_bearer_policy_redirect_customize_sensitive_headers():
    class MockTransport(HttpTransport):
        def __init__(self):
            self._first = True

        def __exit__(self, exc_type, exc_val, exc_tb):
            pass

        def close(self):
            pass

        def open(self):
            pass

        def send(self, request, **kwargs):  # type: (PipelineRequest, Any) -> PipelineResponse
            if self._first:
                self._first = False
                assert request.headers["Authorization"] == "Bearer {}".format(auth_headder)
                response = Response()
                response.status_code = 301
                response.headers["location"] = "https://localhost1"
                return response
            assert request.headers.get("Authorization")
            response = Response()
            response.status_code = 200
            return response

    auth_headder = "token"
    expected_scope = "scope"
    token = AccessToken(auth_headder, 0)
    credential = Mock(get_token=Mock(return_value=token))
    auth_policy = BearerTokenCredentialPolicy(credential, expected_scope)
    redirect_policy = RedirectPolicy()
    header_clean_up_policy = SensitiveHeaderCleanupPolicy(blocked_redirect_headers=["x-ms-authorization-auxiliary"])
    pipeline = Pipeline(transport=MockTransport(), policies=[redirect_policy, auth_policy, header_clean_up_policy])

    pipeline.run(HttpRequest("GET", "https://localhost"))


@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
def test_azure_http_credential_policy(http_request):
    """Tests to see if we can create an AzureHttpKeyCredentialPolicy"""

    prefix = "SharedAccessKey"
    api_key = "test_key"
    header_content = f"{prefix} {api_key}"

    def verify_authorization_header(request):
        assert request.headers["Authorization"] == header_content

    transport = Mock(send=verify_authorization_header)
    credential = AzureKeyCredential(api_key)
    credential_policy = AzureKeyCredentialPolicy(credential=credential, name="Authorization", prefix=prefix)
    pipeline = Pipeline(transport=transport, policies=[credential_policy])

    pipeline.run(http_request("GET", "https://test_key_credential"))



# coding: utf-8
# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE.txt in the project root for
# license information.
# -------------------------------------------------------------------------
import gzip
import tempfile

from flask import (
    Response,
    Blueprint,
)

streams_api = Blueprint("streams_api", __name__)


class StreamingBody:
    def __iter__(self):
        yield b"Hello, "
        yield b"world!"


def streaming_body():
    yield b"Hello, "
    yield b"world!"


def stream_json_error():
    yield '{"error": {"code": "BadRequest", '
    yield ' "message": "You made a bad request"}}'


def streaming_test():
    yield b"test"


def stream_compressed_header_error():
    yield b"test"


@streams_api.route("/basic", methods=["GET"])
def basic():
    return Response(streaming_body(), status=200)


@streams_api.route("/iterable", methods=["GET"])
def iterable():
    return Response(StreamingBody(), status=200)


@streams_api.route("/error", methods=["GET"])
def error():
    return Response(stream_json_error(), status=400)


@streams_api.route("/string", methods=["GET"])
def string():
    return Response(streaming_test(), status=200, mimetype="text/plain")


@streams_api.route("/compressed_no_header", methods=["GET"])
def compressed_no_header():
    return Response(compressed_stream(), status=300)


@streams_api.route("/compressed", methods=["GET"])
def compressed():
    return Response(stream_compressed_header_error(), status=300, headers={"Content-Encoding": "gzip"})


def compressed_stream():
    with tempfile.TemporaryFile(mode="w+b") as f:
        gzf = gzip.GzipFile(mode="w+b", fileobj=f)
        gzf.write(b"test")
        gzf.flush()
        f.seek(0)
        yield f.read()


@streams_api.route("/decompress_header", methods=["GET"])
def decompress_header():
    return Response(compressed_stream(), status=200, headers={"Content-Encoding": "gzip"})



# ------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# ------------------------------------
"""Tests for the user agent policy."""
from azure.core.pipeline.policies import UserAgentPolicy
from azure.core.pipeline import PipelineRequest, PipelineContext

try:
    from unittest import mock
except ImportError:
    import mock
import pytest
from utils import HTTP_REQUESTS


@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
def test_user_agent_policy(http_request):
    user_agent = UserAgentPolicy(base_user_agent="foo")
    assert user_agent._user_agent == "foo"

    user_agent = UserAgentPolicy(sdk_moniker="foosdk/1.0.0")
    assert user_agent._user_agent.startswith("azsdk-python-foosdk/1.0.0 Python")

    user_agent = UserAgentPolicy(base_user_agent="foo", user_agent="bar", user_agent_use_env=False)
    assert user_agent._user_agent == "bar foo"

    request = http_request("GET", "http://localhost/")
    pipeline_request = PipelineRequest(request, PipelineContext(None))

    pipeline_request.context.options["user_agent"] = "xyz"
    user_agent.on_request(pipeline_request)
    assert request.headers["User-Agent"] == "xyz bar foo"


@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
def test_user_agent_environ(http_request):

    with mock.patch.dict("os.environ", {"AZURE_HTTP_USER_AGENT": "mytools"}):
        policy = UserAgentPolicy(None)
        assert policy.user_agent.endswith("mytools")

        request = http_request("GET", "http://localhost/")
        policy.on_request(PipelineRequest(request, PipelineContext(None)))
        assert request.headers["user-agent"].endswith("mytools")



# --------------------------------------------------------------------------
#
# Copyright (c) Microsoft Corporation. All rights reserved.
#
# The MIT License (MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the ""Software""), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.
#
# --------------------------------------------------------------------------
from azure.core.pipeline import policies
from azure.core.configuration import Configuration
from azure.core import PipelineClient
from copy import deepcopy


class TestRestClientConfiguration(Configuration):
    def __init__(self, **kwargs):
        # type: (...) -> None
        super(TestRestClientConfiguration, self).__init__(**kwargs)

        kwargs.setdefault("sdk_moniker", "autorestswaggerbatfileservice/1.0.0b1")
        self._configure(**kwargs)

    def _configure(self, **kwargs):
        # type: (...) -> None
        self.user_agent_policy = kwargs.get("user_agent_policy") or policies.UserAgentPolicy(**kwargs)
        self.headers_policy = kwargs.get("headers_policy") or policies.HeadersPolicy(**kwargs)
        self.proxy_policy = kwargs.get("proxy_policy") or policies.ProxyPolicy(**kwargs)
        self.logging_policy = kwargs.get("logging_policy") or policies.NetworkTraceLoggingPolicy(**kwargs)
        self.http_logging_policy = kwargs.get("http_logging_policy") or policies.HttpLoggingPolicy(**kwargs)
        self.retry_policy = kwargs.get("retry_policy") or policies.RetryPolicy(**kwargs)
        self.custom_hook_policy = kwargs.get("custom_hook_policy") or policies.CustomHookPolicy(**kwargs)
        self.redirect_policy = kwargs.get("redirect_policy") or policies.RedirectPolicy(**kwargs)
        self.authentication_policy = kwargs.get("authentication_policy")


class MockRestClient(object):
    def __init__(self, port, **kwargs):
        self._config = TestRestClientConfiguration(**kwargs)
        self._client = PipelineClient(base_url="http://localhost:{}/".format(port), config=self._config, **kwargs)

    def send_request(self, request, **kwargs):
        """Runs the network request through the client's chained policies.
        >>> from azure.core.rest import HttpRequest
        >>> request = HttpRequest("GET", "http://localhost:3000/helloWorld")
        <HttpRequest [GET], url: 'http://localhost:3000/helloWorld'>
        >>> response = client.send_request(request)
        <HttpResponse: 200 OK>
        For more information on this code flow, see https://aka.ms/azsdk/python/protocol/quickstart
        :param request: The network request you want to make. Required.
        :type request: ~azure.core.rest.HttpRequest
        :keyword bool stream: Whether the response payload will be streamed. Defaults to False.
        :return: The response of your network call. Does not do error handling on your response.
        :rtype: ~azure.core.rest.HttpResponse
        """
        request_copy = deepcopy(request)
        request_copy.url = self._client.format_url(request_copy.url)
        return self._client.send_request(request_copy, **kwargs)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE.txt in the project root for
# license information.
# -------------------------------------------------------------------------

# NOTE: These tests are heavily inspired from the httpx test suite: https://github.com/encode/httpx/tree/master/tests
# Thank you httpx for your wonderful tests!
import pytest
from azure.core.rest import HttpRequest
import collections.abc


@pytest.fixture
def assert_aiterator_body():
    async def _comparer(request, final_value):
        parts = []
        async for part in request.content:
            parts.append(part)
        content = b"".join(parts)
        assert content == final_value

    return _comparer


def test_transfer_encoding_header():
    async def streaming_body(data):
        yield data  # pragma: nocover

    data = streaming_body(b"test 123")

    request = HttpRequest("POST", "http://example.org", data=data)
    assert "Content-Length" not in request.headers


def test_override_content_length_header():
    async def streaming_body(data):
        yield data  # pragma: nocover

    data = streaming_body(b"test 123")
    headers = {"Content-Length": "0"}

    request = HttpRequest("POST", "http://example.org", data=data, headers=headers)
    assert request.headers["Content-Length"] == "0"


@pytest.mark.asyncio
async def test_aiterable_content(assert_aiterator_body):  # cspell:disable-line
    class Content:
        async def __aiter__(self):
            yield b"test 123"

    request = HttpRequest("POST", "http://example.org", content=Content())
    assert request.headers == {}
    await assert_aiterator_body(request, b"test 123")


@pytest.mark.asyncio
async def test_aiterator_content(assert_aiterator_body):
    async def hello_world():
        yield b"Hello, "
        yield b"world!"

    request = HttpRequest("POST", url="http://example.org", content=hello_world())
    assert not isinstance(request._data, collections.abc.Iterable)
    assert isinstance(request._data, collections.abc.AsyncIterable)

    assert request.headers == {}
    await assert_aiterator_body(request, b"Hello, world!")

    # Support 'data' for compat with requests.
    request = HttpRequest("POST", url="http://example.org", data=hello_world())
    assert not isinstance(request._data, collections.abc.Iterable)
    assert isinstance(request._data, collections.abc.AsyncIterable)

    assert request.headers == {}
    await assert_aiterator_body(request, b"Hello, world!")

    # transfer encoding should not be set for GET requests
    request = HttpRequest("GET", url="http://example.org", data=hello_world())
    assert not isinstance(request._data, collections.abc.Iterable)
    assert isinstance(request._data, collections.abc.AsyncIterable)

    assert request.headers == {}
    await assert_aiterator_body(request, b"Hello, world!")


@pytest.mark.asyncio
async def test_read_content(assert_aiterator_body):
    async def content():
        yield b"test 123"

    request = HttpRequest("POST", "http://example.org", content=content())
    await assert_aiterator_body(request, b"test 123")
    # in this case, request._data is what we end up passing to the requests transport
    assert isinstance(request._data, collections.abc.AsyncIterable)
